'''
Created on 2015/11/18

:author: hubo
'''
from vlcp.config import defaultconfig
from vlcp.server.module import Module, api, depend
import vlcp.service.connection.httpserver
from vlcp.utils.http import Dispatcher, HttpInputException
from vlcp.event.stream import MemoryStream, FileStream
from time import time
import os.path
import sys
import logging
from vlcp.event.runnable import RoutineContainer
try:
    from urlparse import urlsplit, urljoin
except:
    from urllib.parse import urlsplit, urljoin
import re
import mimetypes
from email.utils import quote
import stat

withweight = re.compile(br'^([^\s;]+)\s*(?:;\s*[qQ]=((?:1(?:\.0{0,3})?)|(?:0(?:\.[0-9]{0,3})?)))?$')
rangeformat = re.compile(br'([0-9]*)-([0-9]*)')
statusname = re.compile(r'[1-5][0-9]{2}')

def _parseacceptencodings(env):
    elist = env.headerdict.get(b'accept-encoding')
    if elist is None:
        return []
    encoding_list = [e.strip() for e in elist.split(b',')]
    encodings = []
    for enc in encoding_list:
        m = withweight.match(enc)
        if not m:
            return ()
        encodings.append((-1.0 if m.group(2) is None else -float(m.group(2)), m.group(1)))
    allows = []
    identityq = 0
    for q, enc in sorted(encodings):
        if q >= 0:
            # 0 means not acceptable
            break
        enc = enc.lower()
        if enc in allows:
            continue
        if enc == b'identity':
            identityq = q
        elif q > identityq:
            # Ignore encoding with q lower than identity, because identity is always available
            break
        else:
            allows.append(enc)
    return allows
def _parseetag(etag):
    # Ignore whether it is a weak tag
    if etag[:2] == b'W/':
        etag = etag[2:]
    if etag[:1] == b'"' and etag[-1:] == b'"':
        return etag[1:-1]
    else:
        raise ValueError
def _checkrange(env, etag, size):
    if b'range' in env.headerdict:
        if b'if-range' in env.headerdict:
            try:
                if _parseetag(env.headerdict[b'if-range']) != etag:
                    return None
            except ValueError:
                return None
        rangeiden = env.headerdict[b'range']
        # Ignore unrecognized range identifier
        if rangeiden[:6] != b'bytes=':
            return None
        ranges = [r.strip() for r in rangeiden[6:].split(b',')]
        # Accept the first range
        r = ranges[0]
        m = rangeformat.match(r)
        if m is None:
            return None
        if not m.group(1):
            if not m.group(2):
                return None
            else:
                begin = size - int(m.group(2))
                end = size
        else:
            if not m.group(2):
                begin = int(m.group(1))
                end = size
            else:
                # 0-0 means 0:1
                begin = int(m.group(1))
                end = int(m.group(2)) + 1
        # Slice the range to match the correct size
        if begin < 0:
            begin = 0
        if begin > size:
            begin = size
        if end < begin:
            end = begin
        if end > size:
            end = size
        # Ignore invalid ranges
        if end <= begin:
            return None
        return (begin, end)
    else:
        return None

def topath(dirname, tail = None):
    if dirname[:1] != b'/':
        dirname = b'/' + dirname
    if tail:
        if dirname[-1:] != b'/':
            dirname = dirname + b'/'
        dirname = dirname + tail
    return dirname

def _checketag(env, etag):
    nonmatch = env.headerdict.get(b'if-none-match', None)
    if not nonmatch:
        return False
    if nonmatch == b'*':
        return True
    try:
        return etag in [_parseetag(e.strip()) for e in nonmatch.split(b',')]
    except ValueError:
        return False

def safehex(v):
    h = hex(v)[2:]
    if h.endswith('L'):
        h = h[:-1]
    return h

def _createetag(stat_info):
    etag = 'vlcp-' + safehex(stat_info.st_ino) + '-' + safehex(int(stat_info.st_mtime)) + '-' + safehex(int(stat_info.st_size))
    if not isinstance(etag, bytes):
        return etag.encode('ascii')
    return etag

def _generaterange(env, rng, size):
    env.header('Content-Range', 'bytes %d-%d/%d' % (rng[0], rng[1] - 1, size))

@defaultconfig
@depend(vlcp.service.connection.httpserver.HttpServer)
class Static(Module):
    "Map specified path to local files"
    # Not a service
    service = False
    # Check HTTP Referer header to protect against external site links.
    # The Referer header must present and match allowed sites
    _default_checkreferer = False
    # Grant access when Referer header match Host header
    _default_refererallowlocal = True
    # Set allowed referers
    _default_refererallows = []
    # Respond to Range header request
    _default_allowrange = True
    # Generate ETag for static resources, so the browser can use a If-Not-Match request
    # to save bandwidth
    _default_etag = True
    # If there is a "xxxx.gz" file in the folder, it would be recognized as a "xxxx" file
    # with "Content-Encoding: gzip". If gzip = False, it would be recognized as a normal file
    _default_gzip = True
    # Cache expiration time
    _default_maxage = 5
    # Enable a memory cache for small files to improve performance
    _default_memorycache = True
    # Maximum file size allowed to be cached in memory
    _default_memorycachelimit = 4096
    # Maximum cached file number. When it is exceeded, old caches will be cleared.
    _default_memorycacheitemlimit = 4096
    # The directory name for the static resource folder.
    # 
    # Static module supports multiple directories with different configurations, just
    # like vHosts. Use .vdir node to create separated configurations for different
    # folders, e.g.::
    #
    #     module.static.vdir.rewrite.dir="rewrite"
    #     module.static.vdir.rewrite.rewriteonly=True
    # 
    # You may also specify multiple directories with .dirs configuration instead, e.g.::
    #
    #     module.static.dir=None
    #     module.static.dirs=["js","css","images","download"]
    #
    # The directory name can be a relative path against a Python module file, or an absolute path.
    # 
    # .dir and .dirs maps a HTTP GET/HEAD request with path /*dir*/*subpath* to disk file
    # *relativeroot*/*dir*/*subpath*. The *subpath* may be a file in sub directories.
    # ".", ".." is also accepted as current folder/parent folder, but it always map to a path inside
    # *relativeroot*/*dir*, which means:
    # 
    #     /*dir*/*subpath*/a         =>     *relativeroot*/*dir*/a
    #     /*dir*/*subpath*/b/a       =>     *relativeroot*/*dir*/b/a
    #     /*dir*/*subpath*/b/../a    =>     *relativeroot*/*dir*/a
    #     /*dir*/*subpath*/b/../../a =>     *relativeroot*/*dir*/a
    #     
    # So it is not possible to use a static file map to access files outside the mapped folder.
    #
    # If you want to map a HTTP path to a directory with different names, use .map instead
    _default_dir = 'static'
    # Specify the mapped directory is relative to a Python module file. If .relativeroot is also
    # configured and is not empty, .relativeroot takes effect and .relativemodule is ignored.
    # If both are not configured, the current working directory is the relative root.
    # If you start your server with a customized Python script, the default "__main__" will fit.
    # If you start your server with python -m vlcp.start, you should configure the
    # main module manually.
    #
    # Notice that this may import the named Python module.
    _default_relativemodule = '__main__'
    # The relative root of the mapped directory, should be an absolute path if configured.
    _default_relativeroot = None
    # Bind this vdir to a specified HTTP vHost. Create multiple vdirs if you need to provide
    # static file services for multiple HTTP servers.
    _default_vhostbind = ''
    # Bind this vdir to a specified HTTP host, so only requests with corresponding "Host:" header
    # will be responded
    _default_hostbind = None
    # Customized MIME type configuration files, this should be a list (tuple) of file names.
    # Static module use *mimetypes* for guessing MIME information for static file.
    # See mimetypes (https://docs.python.org/2.7/library/mimetypes.html) for more details
    _default_mimetypes = ()
    # When guessing MIME types, use strict mode
    _default_mimestrict = True
    # Customized map for static files, it is an advanced replacement for .dir and .dirs,
    # but they can work at the same time.
    # 
    # The .map configuration should be a dictionary {*http-path*: *file-path*, ...}
    # where *file-path* may be:
    # 
    #     * a tuple (*directory*, *filename*), where *directory* is a directory name similar
    #       to names in .dir or .dirs, and *filename* is a filename or subpath. This maps
    #       *http-path* to *directory*/*filename*. *http-path* and *filename* may use
    #       regular expressions, you may use group capture (brackets in regular expressions)
    #       to capture values, and use them in *filename* with \1, \2, etc.
    # 
    #     * a directory name, in which case *http-path*=>*directory* is equal to
    #       *http-path*/(.\*) => (*directory*, "\1")
    _default_map = {}
    # Use a configured Content-Type, instead of guessing MIME types from file names
    _default_contenttype = None
    # This path cannot be directly accessed from HTTP requests; it only accept a request
    # which is rewritten to this path either by configuration or env.rewrite.
    _default_rewriteonly = False
    # Send extra HTTP headers
    _default_extraheaders = []
    # This directory contains customized error pages. The files should be HTML pages which
    # names start with status numbers, like:
    #
    #     400-badrequest.html
    #     403-forbidden.html
    #     404-notfound.html
    #     500-internalerror.html
    #
    # The status number in the file name will be used for the responding status code.
    # You may configure protocol.http.errorrewrite or protocol.http.errorredirect to
    # rewrite or redirect to error pages:
    # 
    #     protocol.http.errorrewrite={404:b'/error/404-notfound.html',
    #                                 400:b'/error/400-badrequest.html',
    #                                 403:b'/error/403-forbidden.html'}
    _default_errorpage = False
    # Use nginx "X-Accel-Redirect" header to handle the static file request. You must put this server
    # behind nginx proxy, and configure nginx correctly
    _default_xaccelredirect = False
    # Redirect to this path. A request to *dir*/*filename* will be redirected to *redirect_root*/*filename*
    _default_xaccelredirect_root = b'/static'
    # Use Apache X-Sendfile function to handle the static file request. You must put this server
    # behind Apache proxy and configure Apache correctly.
    _default_xsendfile = False
    # Use lighttpd X-LIGHTTPD-send-file to handle the static file request. Newer versions of lighttpd server
    # uses X-Sendfile, so you should set xsendfile=True instead. You must put this server behind lighttpd
    # and configure it correctly.
    _default_xlighttpdsendfile = False
    # Should be None, "attachment", or "inline". If None, "Content-Disposition" header is not used.
    # If set as "attachment", this will usually open a "save as" dialog in browser to let user download
    # the file. If set as "inline", it is processed as normal, but the real file name is sent by
    # the HTTP header, so when user chooses "save as" from browser to save the content,
    # the real file name is used
    _default_contentdisposition = None
    _logger = logging.getLogger(__name__ + '.Static')
    def _clearcache(self, currenttime):
        if self.memorycacheitemlimit <= 0:
            self._cache = {}
            return False
        if currenttime - self.lastcleartime < 1.0:
            # Do not clear too often
            return False
        while len(self._cache) >= self.memorycacheitemlimit:
            del self._cache[min(self._cache.items(), key=lambda x: x[1][2])[0]]
        self.lastcleartime = currenttime
        return True
    def _handlerConfig(self, expand, relativeroot, checkreferer, refererallowlocal, refererallows,
                                           allowrange, etag, gzip, maxage, memorycache,
                                           memorycachelimit, contenttype, rewriteonly, extraheaders,
                                           errorpage, contentdisposition, mimestrict, xaccelredirect,
                                           xsendfile, xlighttpdsendfile, xaccelredirect_root):
        def handler(env):
            currenttime = time()
            if rewriteonly:
                if not env.rewritefrom:
                    for m in env.error(404):
                        yield m
                    env.exit()
            if not errorpage and checkreferer:
                try:
                    referer = env.headerdict.get(b'referer')
                    if referer is None:
                        referer_host = None
                    else:
                        referer_host = urlsplit(referer).netloc
                    if not ((refererallowlocal and referer_host == env.host) or
                        referer_host in refererallows):
                        for m in env.error(403, showerror = False):
                            yield m
                        env.exit()
                except:
                    for m in env.error(403, showerror = False):
                        yield m
                    env.exit()
            localpath = env.path_match.expand(expand)
            realpath = env.getrealpath(relativeroot, localpath)
            filename = os.path.basename(realpath)
            if xsendfile or xlighttpdsendfile or xaccelredirect:
                # Apache send a local file
                env.startResponse(200)
                if contenttype:
                    env.header('Content-Type', contenttype)
                else:
                    mime = self.mimetypedatabase.guess_type(filename, mimestrict)
                    if mime[1]:
                        # There should not be a content-encoding here, maybe the file itself is compressed
                        # set mime to application/octet-stream
                        mime_type = 'application/octet-stream'
                    elif not mime[0]:
                        mime_type = 'application/octet-stream'
                    else:
                        mime_type = mime[0]
                    env.header('Content-Type', mime_type, False)
                if not errorpage and contentdisposition:
                    env.header('Content-Disposition', contentdisposition + '; filename=' + quote(filename))
                if xsendfile:
                    env.header('X-Sendfile', realpath)
                if xaccelredirect:
                    env.header(b'X-Accel-Redirect', urljoin(xaccelredirect_root, self.dispatcher.expand(env.path_match, expand)))
                if xlighttpdsendfile:
                    env.header(b'X-LIGHTTPD-send-file', realpath)
                env.exit()
                
            use_gzip = False
            if gzip:
                if realpath.endswith('.gz'):
                    # GZIP files are preserved for gzip encoding
                    for m in env.error(403, showerror = False):
                        yield m
                    env.exit()
                encodings = _parseacceptencodings(env)
                if b'gzip' in encodings or b'x-gzip' in encodings:
                    use_gzip = True
            use_etag = etag and not errorpage
            # First time cache check
            if memorycache:
                # Cache data: (data, headers, cachedtime, etag)
                cv = self._cache.get((realpath, use_gzip))
                if cv and cv[2] + max(0 if maxage is None else maxage, 3) > currenttime:
                    # Cache is valid
                    if use_etag:
                        if _checketag(env, cv[3]):
                            env.startResponse(304, cv[1])
                            env.exit()
                    size = len(cv[0])
                    rng = None
                    if not errorpage and allowrange:
                        rng = _checkrange(env, cv[3], size)
                    if rng is not None:
                        env.startResponse(206, cv[1])
                        _generaterange(env, rng, size)
                        env.output(MemoryStream(cv[0][rng[0]:rng[1]]), use_gzip)
                    else:
                        if errorpage:
                            m = statusname.match(filename)
                            if m:
                                env.startResponse(int(m.group()), cv[1])
                            else:
                                # Show 200-OK is better than 500
                                env.startResponse(200, cv[1])
                        else:
                            env.startResponse(200, cv[1])
                        env.output(MemoryStream(cv[0]), use_gzip)
                    env.exit()
            # Test file
            if use_gzip:
                try:
                    stat_info = os.stat(realpath + '.gz')
                    if not stat.S_ISREG(stat_info.st_mode):
                        raise ValueError('Not regular file')
                    realpath += '.gz'
                except:
                    try:
                        stat_info = os.stat(realpath)
                        if not stat.S_ISREG(stat_info.st_mode):
                            raise ValueError('Not regular file')
                        use_gzip = False
                    except:
                        for m in env.error(404, showerror = False):
                            yield m
                        env.exit()
            else:
                try:
                    stat_info = os.stat(realpath)
                    if not stat.S_ISREG(stat_info.st_mode):
                        raise ValueError('Not regular file')
                    use_gzip = False
                except:
                    for m in env.error(404, showerror = False):
                        yield m
                    env.exit()
            newetag = _createetag(stat_info)
            # Second memory cache test
            if memorycache:
                # use_gzip may change
                cv = self._cache.get((realpath, use_gzip))
                if cv and cv[3] == newetag:
                    # Cache is valid
                    if use_etag:
                        if _checketag(env, cv[3]):
                            env.startResponse(304, cv[1])
                            env.exit()
                    self._cache[(realpath, use_gzip)] = (cv[0], cv[1], currenttime, newetag)
                    size = len(cv[0])
                    rng = None
                    if not errorpage and allowrange:
                        rng = _checkrange(env, cv[3], size)
                    if rng is not None:
                        env.startResponse(206, cv[1])
                        _generaterange(env, rng, size)
                        env.output(MemoryStream(cv[0][rng[0]:rng[1]]), use_gzip)
                    else:
                        if errorpage:
                            m = statusname.match(filename)
                            if m:
                                env.startResponse(int(m.group()), cv[1])
                            else:
                                # Show 200-OK is better than 500
                                env.startResponse(200, cv[1])
                        else:
                            env.startResponse(200, cv[1])
                        env.output(MemoryStream(cv[0]), use_gzip)
                    env.exit()
                elif cv:
                    # Cache is invalid, remove it to prevent another hit
                    del self._cache[(realpath, use_gzip)]
            # No cache available, get local file
            # Create headers
            if contenttype:
                env.header('Content-Type', contenttype)
            else:
                mime = self.mimetypedatabase.guess_type(filename, mimestrict)
                if mime[1]:
                    # There should not be a content-encoding here, maybe the file itself is compressed
                    # set mime to application/octet-stream
                    mime_type = 'application/octet-stream'
                elif not mime[0]:
                    mime_type = 'application/octet-stream'
                else:
                    mime_type = mime[0]
                env.header('Content-Type', mime_type, False)
            if use_etag:
                env.header(b'ETag', b'"' + newetag + b'"', False)
            if maxage is not None:
                env.header('Cache-Control', 'max-age=' + str(maxage), False)
            if use_gzip:
                env.header(b'Content-Encoding', b'gzip', False)
            if not errorpage and contentdisposition:
                env.header('Content-Disposition', contentdisposition + '; filename=' + quote(filename))
            if allowrange:
                env.header(b'Accept-Ranges', b'bytes')
            if extraheaders:
                env.sent_headers.extend(extraheaders)
            if use_etag:
                if _checketag(env, newetag):
                    env.startResponse(304, clearheaders = False)
                    env.exit()
            if memorycache and stat_info.st_size <= memorycachelimit:
                # Cache
                cache = True
                if len(self._cache) >= self.memorycacheitemlimit:
                    if not self._clearcache(currenttime):
                        cache = False
                if cache:
                    with open(realpath, 'rb') as fobj:
                        data = fobj.read()
                    self._cache[(realpath, use_gzip)] = (data, env.sent_headers[:], currenttime, newetag)
                    size = len(data)
                    rng = None
                    if not errorpage and allowrange:
                        rng = _checkrange(env, newetag, size)
                    if rng is not None:
                        env.startResponse(206, clearheaders = False)
                        _generaterange(env, rng, size)
                        env.output(MemoryStream(data[rng[0]:rng[1]]), use_gzip)
                    else:
                        if errorpage:
                            m = statusname.match(filename)
                            if m:
                                env.startResponse(int(m.group()), clearheaders = False)
                            else:
                                # Show 200-OK is better than 500
                                env.startResponse(200, clearheaders = False)
                        else:
                            env.startResponse(200, clearheaders = False)
                        env.output(MemoryStream(data), use_gzip)
                    env.exit()
            size = stat_info.st_size
            if not errorpage and allowrange:
                rng = _checkrange(env, newetag, size)
            if rng is not None:
                env.startResponse(206, clearheaders = False)
                _generaterange(env, rng, size)
                fobj = open(realpath, 'rb')
                try:
                    fobj.seek(rng[0])
                except:
                    fobj.close()
                    raise
                else:
                    env.output(FileStream(fobj, isunicode=False, size=rng[1] - rng[0]), use_gzip)
            else:
                if errorpage:
                    m = statusname.match(filename)
                    if m:
                        env.startResponse(int(m.group()), clearheaders = False)
                    else:
                        # Show 200-OK is better than 500
                        env.startResponse(200, clearheaders = False)
                else:
                    env.startResponse(200, clearheaders = False)
                env.output(FileStream(open(realpath, 'rb'), isunicode = False), use_gzip)
        return handler
    _configurations = ['checkreferer', 'refererallowlocal', 'refererallows',
            'allowrange', 'etag', 'gzip', 'maxage', 'memorycache',
            'memorycachelimit', 'contenttype', 'rewriteonly', 'extraheaders',
            'errorpage', 'contentdisposition', 'mimestrict', 'xaccelredirect',
            'xsendfile', 'xlighttpdsendfile', 'xaccelredirect_root']
    def _createHandlers(self, config, defaultconfig = {}):
        dirs = list(getattr(config, 'dirs', []))
        if hasattr(config, 'dir') and config.dir and config.dir.strip():
            dirs.append(config.dir)
        # Change to str
        dirs = [d.decode('utf-8') if not isinstance(d, str) else d for d in dirs]
        maps = dict((topath(d.encode('utf-8'), b'(.*)'), (d, br'\1')) for d in dirs)
        if hasattr(config, 'map') and config.map:
            for k,v in config.map.items():
                if not isinstance(k, bytes):
                    k = k.encode('utf-8')
                if isinstance(v, str) or isinstance(v, bytes):
                    if not isinstance(v, str):
                        v = v.decode('utf-8')
                    # (b'/abc' => 'def') is equal to (b'/abc/(.*)' => ('def', br'\1')
                    maps[topath(k, b'(.*)')] = (v, br'\1')
                else:
                    # Raw map
                    # (b'/' => (b'static', b'index.html'))
                    d = v[0]
                    if not isinstance(d, str):
                        d = d.decode('utf-8')
                    expand = v[1]
                    if not isinstance(d, bytes):
                        expand = expand.encode('utf-8')
                    maps[topath(k)] = (d, expand)
        getconfig = lambda k: getattr(config, k) if hasattr(config, k) else defaultconfig.get(k)
        hostbind = getconfig('hostbind')
        vhostbind = getconfig('vhostbind')
        relativeroot = getconfig('relativeroot')
        relativemodule = getconfig('relativemodule')
        if maps:
            # Create configuration
            newconfig = dict((k,getconfig(k)) for k in self._configurations)
            if not relativeroot:
                if relativemodule:
                    try:
                        __import__(relativemodule)
                        mod = sys.modules[relativemodule]
                        filepath = getattr(mod, '__file__', None)
                        if filepath:
                            relativeroot = os.path.dirname(filepath)
                        else:
                            self._logger.warning('Relative module %r has no __file__, use cwd %r instead',
                                                 relativemodule,
                                                 os.getcwd())
                            relativeroot = os.getcwd()
                    except:
                        self._logger.exception('Cannot locate relative module %r', relativemodule)
                        raise
                else:
                    relativeroot = os.getcwd()
            relativeroot = os.path.abspath(relativeroot)
            for k,v in maps.items():
                drel = os.path.normpath(os.path.join(relativeroot, v[0]))
                if not os.path.isdir(drel):
                    self._logger.error('Cannot find directory: %r', drel)
                    continue
                # For security reason, do not allow a package directory to be exported
                if os.path.isfile(os.path.join(drel, '__init__.py')):
                    self._logger.error('Path %r is a package', drel)
                    continue
                self.dispatcher.route(k, self._handlerConfig(v[1], drel, **newconfig), self.apiroutine,
                                      hostbind, vhostbind)
        if hasattr(config, 'vdir'):
            newconfig.update((('hostbind', hostbind), ('vhostbind', vhostbind),
                              ('relativeroot', getconfig('relativeroot')),('relativemodule', relativemodule)))
            for k,v in config.vdir.items():
                self._createHandlers(v, newconfig)
    def __init__(self, server):
        Module.__init__(self, server)
        self.dispatcher = Dispatcher(self.scheduler)
        self.mimetypedatabase = mimetypes.MimeTypes(self.mimetypes)
        self._cache = {}
        self.apiroutine = RoutineContainer(self.scheduler)
        self.lastcleartime = 0
        def start(asyncStart = False):
            self._createHandlers(self)
        def close():
            self.dispatcher.close()
        self.apiroutine.start = start
        self.apiroutine.close = close
        self.routines.append(self.apiroutine)
        self.createAPI(api(self.updateconfig))
    def updateconfig(self):
        "Reload configurations, remove non-exist servers, add new servers, and leave others unchanged"
        self.dispatcher.unregisterAllHandlers()
        self._createHandlers(self)
        return None



'''
Created on 2015/11/9

:author: hubo
'''
from vlcp.config import defaultconfig
from vlcp.server.module import Module, api, depend, callAPI
from vlcp.event.runnable import RoutineContainer
from vlcp.event import Event, withIndices
from vlcp.service.utils import knowledge
from uuid import uuid4
try:
    from Cookie import SimpleCookie, Morsel
except:
    from http.cookies import SimpleCookie, Morsel
from vlcp.event.lock import Lock

@depend(knowledge.MemoryStorage)
@defaultconfig
class Session(Module):
    '''
    HTTP Session with cookies
    '''
    # Timeout for a session
    _default_timeout = 1800
    # Cookie name used in session id storage
    _default_cookiename = '_session_id'
    service = True
    class SessionObject(object):
        def __init__(self, sessionid):
            self.id = sessionid
            self.vars = {}
            self.lockseq = 0
            self.releaseseq = 0
    class SessionHandle(object):
        def __init__(self, sessionobj, container):
            self.sessionobj = sessionobj
            self.id = sessionobj.id
            self.vars = sessionobj.vars
            self._lock = Lock(sessionobj, container.scheduler)
            self.container = container
        def lock(self):
            "Lock session"
            for m in self._lock.lock(self.container):
                yield m
        def unlock(self):
            "Unlock session"
            self._lock.unlock()
    def __init__(self, server):
        '''
        Constructor
        '''
        Module.__init__(self, server)
        self.apiroutine = RoutineContainer(self.scheduler)
        self.createAPI(api(self.start, self.apiroutine),
                       api(self.create, self.apiroutine),
                       api(self.get, self.apiroutine),
                       api(self.destroy, self.apiroutine))
    def start(self, cookies, cookieopts = None):
        """
        Session start operation. First check among the cookies to find existed sessions;
        if there is not an existed session, create a new one.
        
        :param cookies: cookie header from the client
        
        :param cookieopts: extra options used when creating a new cookie
        
        :return: ``(session_handle, cookies)`` where session_handle is a SessionHandle object,
                 and cookies is a list of created Set-Cookie headers (may be empty)
        """
        c = SimpleCookie(cookies)
        sid = c.get(self.cookiename)
        create = True
        if sid is not None:
            for m in self.get(sid.value):
                yield m
            if self.apiroutine.retvalue is not None:
                self.apiroutine.retvalue = (self.SessionHandle(self.apiroutine.retvalue, self.apiroutine), [])
                create = False
        if create:
            for m in self.create():
                yield m
            sh = self.apiroutine.retvalue
            m = Morsel()
            m.key = self.cookiename
            m.value = sh.id
            m.coded_value = sh.id
            opts = {'path':'/', 'httponly':True}
            if cookieopts:
                opts.update(cookieopts)
                if not cookieopts['httponly']:
                    del cookieopts['httponly']
            m.update(opts)
            self.apiroutine.retvalue = (sh, [m])
    def get(self, sessionid, refresh = True):
        """
        Get the seesion object of the session id
        
        :param sessionid: a session ID
        
        :param refresh: if True, refresh the expiration time of this session
        
        :return: Session object or None if not exists
        """
        for m in callAPI(self.apiroutine, 'memorystorage', 'get', {'key': __name__ + '.' + sessionid, 'timeout': self.timeout if refresh else None}):
            yield m
    def create(self):
        """
        Create a new session object
        
        :return: Session handle for the created session object.
        """
        sid = uuid4().hex
        sobj = self.SessionObject(sid)
        for m in callAPI(self.apiroutine, 'memorystorage', 'set', {'key': __name__ + '.' + sid, 'value': sobj, 'timeout': self.timeout}):
            yield m
        self.apiroutine.retvalue = self.SessionHandle(sobj, self.apiroutine)
    def destroy(self, sessionid):
        """
        Destroy a session
        
        :param sessionid: session ID
        
        :return: a list of Set-Cookie headers to be sent to the client
        """
        for m in callAPI(self.apiroutine, 'memorystorage', 'delete', {'key': __name__ + '.' + sessionid}):
            yield m
        m = Morsel()
        m.key = self.cookiename
        m.value = 'deleted'
        m.coded_value = 'deleted'
        opts = {'path':'/', 'httponly':True, 'max-age':0}
        m.update(opts)
        self.apiroutine.retvalue = [m]
    


'''
Created on 2015/12/2

:author: hubo
'''

from vlcp.config import defaultconfig
from vlcp.server.module import Module, api, depend, callAPI
from vlcp.event.core import TimerEvent
from vlcp.event.runnable import RoutineContainer
import vlcp.service.connection.httpserver
from time import time
import os.path
from vlcp.utils.http import HttpHandler
from email.message import Message
import json
import ast
from namedstruct import NamedStruct, dump
from vlcp.utils.jsonencoder import encode_default, decode_object, JsonFormat
import traceback

def _str(b, encoding = 'ascii'):
    if isinstance(b, str):
        return b
    else:
        return b.decode(encoding)

class WebAPIHandler(HttpHandler):
    def __init__(self, parent):
        HttpHandler.__init__(self, scheduler=parent.scheduler, daemon=False, vhost=parent.vhostbind)
        self.parent = parent
        self.jsonencoder = JsonFormat()
    def apiHandler(self, env, targetname, methodname, **kwargs):
        params = kwargs
        parent = self.parent
        if not params:
            if b'content-type' in env.headerdict and env.inputstream is not None and parent.acceptjson:
                m = Message()
                m['content-type'] = _str(env.headerdict[b'content-type'])
                if m.get_content_type() == 'application/json':
                    charset = m.get_content_charset('utf-8')
                    for m in env.inputstream.read(self):
                        yield m
                    params = json.loads(_str(self.data, charset), encoding=charset, object_hook=decode_object)
        elif parent.typeextension:
            for k in params.keys():
                v = params[k]
                if v[:1] == '`' and v[-1:] == '`':
                    try:
                        params[k] = ast.literal_eval(v[1:-1])
                    except:
                        pass
        if parent.allowtargets is not None:
            if targetname not in parent.allowtargets:
                for m in env.error(403):
                    yield m
                return
        elif parent.denytargets is not None:
            if targetname in parent.denytargets:
                for m in env.error(403):
                    yield m
                return
        if parent.authmethod:
            for m in callAPI(self, parent.authtarget, parent.authmethod,
                             {'env':env, 'targetname':targetname, 'name':methodname, 'params': params}):
                yield m
        try:
            for m in callAPI(self, targetname, methodname, params):
                yield m
        except Exception as exc:
            if parent.errordetails:
                parent._logger.warning('Web API call failed for %r/%r', targetname, methodname, exc_info = True)
                env.startResponse(500, [(b'Content-Type', b'application/json')])
                err = {'error': str(exc)}
                if parent.errortrace:
                    err['trace'] = traceback.format_exc()
                env.outputjson(err)
            else:
                raise
        else:
            env.header('Content-Type', 'application/json')
            env.outputdata(json.dumps({'result':self.retvalue}, default=self.jsonencoder.jsonencoder).encode('ascii'))
    def start(self, asyncStart=False):
        HttpHandler.start(self, asyncStart=asyncStart)
        path = self.parent.rootpath.encode('utf-8')
        if path[-1:] != b'/':
            path += b'/'
        path += b'(?P<targetname>[^/]*)/(?P<methodname>[^/]*)'
        self.dispatcher.routeargs(path, self.apiHandler, self, self.parent.hostbind,
                                  self.parent.vhostbind, self.parent.acceptmethods,
                                  matchargs = ('targetname', 'methodname'), csrfcheck = False)
@depend(vlcp.service.connection.httpserver.HttpServer)
@defaultconfig
class WebAPI(Module):
    '''
    Call module API from web. Free access to any module APIs may create serious security problems,
    make sure to configure this module properly.
    '''
    # Default bind HttpServer vHost to 'api', so you should use::
    # 
    #     module.httpserver.vhost.api.url='http://localhost:8080/'
    # 
    # to create the API endpoint
    _default_vhostbind = 'api'
    # Bind to a specified Host (HTTP "Host: " header)
    _default_hostbind = None
    # Bind API endpoint under this path. Each endpoint would be \<rootpath\>/\<targetname\>/\<methodname\>.
    # e.g. Manager.reloadmodules would be::
    # 
    #     http://<serverurl>/<rootpath>/manager/reloadmodules
    # 
    # targetname and methodnames are always in lower case
    _default_rootpath = '/'
    # Allowed HTTP method, GET/POST or both
    _default_acceptmethods = [b'GET', b'POST']
    # Allow to use JSON format for POST data
    _default_acceptjson = True
    # If authenticate is enabled, WebAPI module uses an extra API call to authenticate the request.
    # This is the target name.
    _default_authtarget = 'public'
    # If authenticate is enabled, WebAPI module uses an extra API call to authenticate the request.
    # This is the method name. The params will be {'env': *env*, 'targetname':*targetname*,
    # 'name': *methodname*, 'params': *parameters*}
    _default_authmethod = None
    # Only allow API calls on specified targets (usually a target name is the lower-cased module name)
    _default_allowtargets = None
    # Disallow API calls on specified targets
    _default_denytargets = None
    # Enable a type extension for GET/POST methods: use `` quoted string to represent a Python literal
    # expression like lists, tuples, dictionaries, numbers etc.
    _default_typeextension = True
    # Show error details in error responses with JSON format {"error":...}. If errordetails = False,
    # the default HTTP error response is used.
    _default_errordetails = True
    # Also output the trace information in the JSON output
    _default_errortrace = False
    service = False
    def __init__(self, server):
        '''
        Constructor
        '''
        Module.__init__(self, server)
        self.routines.append(WebAPIHandler(self))


'''
Created on 2016/3/25

:author: think
'''

from vlcp.service.utils.knowledge import escape_key, unescape_key
from vlcp.event.event import withIndices, Event
from contextlib import contextmanager
from vlcp.server.module import callAPI
import functools
from copy import deepcopy
from vlcp.event.core import QuitException
import itertools
try:
    from itertools import izip
except ImportError:
    def izip(*args, **kwargs):
        return iter(zip(*args, **kwargs))

@withIndices('key', 'transactid', 'type')
class DataObjectUpdateEvent(Event):
    UPDATED = 'updated'
    DELETED = 'deleted'

class ReferenceObject(object):
    "A strong reference. The referenced object should be automatically retrieved from KVDB."
    def __init__(self, key, refobj = None):
        self._key = key
        self._ref = refobj
    def getkey(self):
        return self._key
    def kvdb_retrievelist(self):
        if self._ref is None:
            return [self.getkey()]
        else:
            return []
    def kvdb_retrievefinished(self, result):
        self._ref = result.get(self._key)
    def kvdb_internalref(self):
        return set((self._key,))
    def kvdb_detach(self):
        pass
    def isdeleted(self):
        return self._ref is None or self._ref._deleted
    def __getattr__(self, key):
        if key[:1] == '_' and not key.startswith('kvdb_'):
            raise AttributeError("type object %r has no attribute %r" % (type(self).__name__, key))
        if self._ref is None:
            raise AttributeError("Target object is not connected")
        return getattr(self._ref, key)
    def jsonencode(self):
        return self._key
    @classmethod
    def jsondecode(cls, data):
        return cls(data, None)
    def __getstate__(self):
        return self._key
    def __setstate__(self, state):
        self._key = state
        self._ref = None
    def __hash__(self, *args, **kwargs):
        return hash(self._key) ^ hash('ReferenceObject')
    def __eq__(self, obj):
        if not isinstance(obj, ReferenceObject):
            return NotImplemented
        else:
            return (obj is self or obj._key == self._key)
    def __ne__(self, obj):
        if not isinstance(obj, ReferenceObject):
            return NotImplemented
        else:
            return not (obj is self or obj._key == self._key)
    def __repr__(self, *args, **kwargs):
        return '<ReferenceObject: %r at %r>' % (self._ref, self._key)
    def __str__(self, *args, **kwargs):
        return self._key
    def wait(self, container):
        if self.isdeleted():
            yield (DataObjectUpdateEvent.createMatcher(self.getkey(), None, DataObjectUpdateEvent.UPDATED),)
            self._ref = container.event.object
    def waitif(self, container, expr, nextchange = False):
        flag = nextchange
        while True:
            r = not flag and expr(self)
            flag = False
            if r:
                container.retvalue = r
                break
            yield (DataObjectUpdateEvent.createMatcher(self.getkey()),)
            if self._ref is not container.event.object:
                self._ref = container.event.object
    
class WeakReferenceObject(object):
    "A weak reference. The referenced object must be retrieved manually."
    def __init__(self, key):
        self._key = key
    def getkey(self):
        return self._key
    def jsonencode(self):
        return self._key
    @classmethod
    def jsondecode(cls, data):
        return cls(data)
    def __getstate__(self):
        return self._key
    def __setstate__(self, state):
        self._key = state
    def __hash__(self, *args, **kwargs):
        return hash(self._key) ^ hash('WeakReferenceObject')
    def __eq__(self, obj):
        if not isinstance(obj, WeakReferenceObject):
            return NotImplemented
        else:
            return (obj is self or obj._key == self._key)
    def __ne__(self, obj):
        if not isinstance(obj, WeakReferenceObject):
            return NotImplemented
        else:
            return not (obj is self or obj._key == self._key)
    def __repr__(self, *args, **kwargs):
        return '<WeakReferenceObject: %r>' % (self._key,)
    def __str__(self, *args, **kwargs):
        return self._key

class DataObject(object):
    "A base class to serialize data into KVDB"
    _prefix = ''
    _indices = ()
    # Unique key and multiple key should like: (('keyname1', ('keyattr1', 'keyattr2', ...),), ('keyname2', ...), ...)
    _unique_keys = ()
    _multi_keys = ()
    _auto_removes = {}
    @classmethod
    def _register_auto_remove(cls, key, func):
        if '_auto_removes' not in cls.__dict__:
            cls._auto_removes = {key: func}
        else:
            cls.__dict__['_auto_removes'][key] = func
    def __init__(self, prefix = None, deleted = False):
        if prefix is not None:
            self._prefix = prefix
        self._deleted = deleted
    def getkey(self):
        if not self._indices:
            return self._prefix
        else:
            return self._prefix + '.' + '.'.join(escape_key(str(getattr(self, ind))) for ind in self._indices)
    @classmethod
    def _getIndices(cls, key):
        if not cls._indices:
            return (key, [])
        keys = key.split('.')
        indices = cls._indices
        if len(keys) < len(indices):
            raise ValueError('Malformed key: indices %r not found' % (indices,))
        else:
            prefix = '.'.join(keys[:-len(indices)])
        return (prefix, [unescape_key(k) for k in keys[-len(indices):]])
    @classmethod
    def ismatch(cls, key, matchvalues):
        _, values = cls._getIndices(key)
        return all(v == '%' or mv is None or v == mv for v, mv in zip(values, matchvalues))
    @classmethod
    def isinstance(cls, cls2):
        # Compare by name
        def clsname(c):
            return c.__module__ + '.' + c.__name__
        try:
            return clsname(cls2) in (clsname(c) for c in cls.__mro__)
        except Exception:
            return isinstance(cls, cls2)
    @classmethod
    def create_instance(cls, *args):
        obj = cls()
        for k,v in zip(obj._indices, args):
            setattr(obj, k, v)
        return obj
    @classmethod
    def create_from_key(cls, key):
        obj = cls()
        obj.setkey(key)
        return obj
    @classmethod
    def default_key(cls, *args):
        if not cls._indices:
            return cls._prefix
        else:
            return cls._prefix + '.' + '.'.join(escape_key(str(a)) for a in args)
    def setkey(self, key):
        self._prefix, values = self._getIndices(key)
        self.__dict__.update(zip(self._indices, values))
    def kvdb_retrievelist(self):
        if self._deleted:
            return []
        return_list = set()
        for k,v in self.__dict__.items():
            if k[:1] != '_' and hasattr(v, 'kvdb_retrievelist'):
                return_list.update(v.kvdb_retrievelist())
        return list(return_list)
    def kvdb_retrievefinished(self, result):
        for k,v in self.__dict__.items():
            if k[:1] != '_' and hasattr(v, 'kvdb_retrievefinished'):
                v.kvdb_retrievefinished(result)
    def jsonencode(self):
        return dict((k,v) for k,v in self.__dict__.items() if k[:1] != '_' and k not in self._indices)
    @classmethod
    def jsondecode(cls, data):
        obj = cls(None, False)
        obj.__dict__.update((str(k),v) for k,v in data.items())
        return obj
    def __getstate__(self):
        return (dict((k,v) for k,v in self.__dict__.items() if k[:1] != '_' and k not in self._indices), False)
    def __setstate__(self, state):
        self.__init__(None, False)
        self.__dict__.update(state[0])
    def clone_instance(self):
        c = deepcopy(self)
        c.setkey(self.getkey())
        return c
    def create_reference(self):
        return ReferenceObject(self.getkey(), self)
    def create_weakreference(self):
        return WeakReferenceObject(self.getkey())
    def __repr__(self, *args, **kwargs):
        return '<DataObject (%s) at %r>' % (object.__repr__(self, *args, **kwargs), self.getkey())
    def __eq__(self, obj):
        if not isinstance(obj, DataObject):
            return NotImplemented
        else:
            return (obj is self or \
                dict((k,v) for k,v in self.__dict__.items() if k[:1] != '_') == dict((k,v) for k,v in obj.__dict__.items() if k[:1] != '_'))
    def __hash__(self):
        raise TypeError('unhashable type: DataObject')
    def __ne__(self, obj):
        r = self.__eq__(obj)
        if r is NotImplemented:
            return r
        else:
            return not r
    def kvdb_update(self, obj):
        for k,v in list(self.__dict__.items()):
            if k[:1] != '_':
                if k in obj.__dict__ and v != obj.__dict__[k]:
                    if hasattr(v, 'kvdb_update'):
                        v.kvdb_update(obj.__dict__[k])
                    else:
                        if hasattr(v, 'kvdb_detach'):
                            v.kvdb_detach()
                        self.__dict__[k] = obj.__dict__[k]
                elif k not in obj.__dict__:
                    if hasattr(v, 'kvdb_detach'):
                        v.kvdb_detach()
                    del self.__dict__[k]
        for k,v in obj.__dict__.items():
            if k not in self.__dict__:
                self.__dict__[k] = v
    def kvdb_detach(self):
        if not self._deleted:
            for k in [k for k in self.__dict__.keys() if k[:1] != '_']:
                if k in self.__dict__ and hasattr(self.__dict__[k], 'kvdb_detach'):
                    self.__dict__[k].kvdb_detach()
            self._deleted = True
    def kvdb_internalref(self):
        r = set()
        for k,v in self.__dict__.items():
            if k[:1] != '_' and v is not None and hasattr(v, 'kvdb_internalref'):
                r.update(v.kvdb_internalref())
        return r
    @classmethod
    def unique_key(cls, keyname, *args):
        return UniqueKeyReference.default_key(cls._prefix, keyname, '.'.join(escape_key(str(a)) for a in args))
    @classmethod
    def multi_key(cls, keyname, *args):
        return MultiKeyReference.default_key(cls._prefix, keyname, '.'.join(escape_key(str(a)) for a in args))
    def kvdb_uniquekeys(self):
        return [self.unique_key(k, *[getattr(self, ind) for ind in indices])
                for k,indices in self._unique_keys
                if all(hasattr(self, ind) for ind in indices)]
    def kvdb_multikeys(self):
        return [self.multi_key(k, *[getattr(self, ind) for ind in indices])
                for k,indices in self._multi_keys
                if all(hasattr(self, ind) for ind in indices)]
    def kvdb_autoremove(self):
        auto_removes = {}
        for c in reversed(type(self).__mro__):
            if '_auto_removes' in c.__dict__:
                auto_removes.update(c.__dict__['_auto_removes'])
        return set(itertools.chain.from_iterable(v(self) for v in auto_removes.values()))
        
class DataObjectSet(object):
    "A set of data objects, usually of a same type. Allow weak references only."
    def __init__(self):
        self._dataset = set()
        self._dataindices = None
        self._lastclass = None
    def jsonencode(self):
        return list(self._dataset)
    @classmethod
    def jsondecode(cls, data):
        obj = cls.__new__(cls)
        obj.__init__()
        obj._dataset = set(data)
        return obj
    def __getstate__(self):
        return (list(self._dataset), True)
    def __setstate__(self, state):
        self.__init__()
        self._dataset = set(state[0])
    def _create_indices(self, cls):
        if self._dataindices is not None and cls is self._lastclass:
            return
        self._dataindices = None
        self._lastclass = cls
        for robj in self._dataset:
            _, values = cls._getIndices(robj.getkey())
            if self._dataindices is None:
                self._dataindices = [{} for _ in range(0, len(values))]
            for i in range(0, len(values)):
                self._dataindices[i].setdefault(values[i], set()).add(robj)
    def find(self, cls, *args):
        self._create_indices(cls)
        if self._dataindices is None:
            # No Data
            return []
        curr = None
        for i in range(0, len(args)):
            if args[i] is not None:
                curr_match = self._dataindices[i].get(args[i], set()).union(self._dataindices[i].get('%', set()))
                if curr is None:
                    curr = curr_match
                else:
                    curr.intersection_update(curr_match)
        if curr is None:
            return list(self._dataset)
        else:
            return list(curr)
    def __repr__(self, *args, **kwargs):
        return '<DataObjectSet %r>' % (self._dataset,)
    def __eq__(self, obj):
        if not isinstance(obj, DataObjectSet):
            return NotImplemented
        else:
            return (obj is self or self._dataset == obj._dataset)
    def __ne__(self, obj):
        if not isinstance(obj, DataObjectSet):
            return NotImplemented
        else:
            return not (obj is self or self._dataset == obj._dataset)
    def __hash__(self):
        raise TypeError('unhashable type: DataObject')
    def kvdb_update(self, obj):
        self._dataset.clear()
        self._dataset.update(obj._dataset)
        self._dataindices = None
        self._lastclass = None
    def dataset(self):
        return self._dataset

class UniqueKeySet(DataObject):
    _prefix = 'indicesset'
    _indices = ('prefix', 'keyname')
    def __init__(self, prefix=None, deleted=False):
        super(UniqueKeySet, self).__init__(prefix=prefix, deleted=deleted)
        self.set = DataObjectSet()


class MultiKeySet(DataObject):
    _prefix = 'multiindicesset'
    _indices = ('prefix', 'keyname')
    def __init__(self, prefix=None, deleted=False):
        super(MultiKeySet, self).__init__(prefix=prefix, deleted=deleted)
        self.set = DataObjectSet()


class UniqueKeyReference(DataObject):
    _prefix = 'indices'
    _indices = ('prefix', 'keyname', 'value')
    @classmethod
    def get_keyset_from_key(cls, key):
        _, (prefix, keyname, _) = cls._getIndices(key)
        return UniqueKeySet.default_key(prefix, keyname)

        
class MultiKeyReference(DataObject):
    _prefix = 'multiindices.'
    _indices = ('prefix', 'keyname', 'value')
    @classmethod
    def get_keyset_from_key(cls, key):
        _, (prefix, keyname, _) = cls._getIndices(key)
        return MultiKeySet.default_key(prefix, keyname)
    

@contextmanager
def watch_context(keys, result, reqid, container, module = 'objectdb'):
    try:
        keys = [k for k,r in zip(keys, result) if r is not None]
        yield result
    finally:
        if keys:
            def clearup():
                try:
                    for m in callAPI(container, module, 'munwatch', {'keys': keys, 'requestid': reqid}):
                        yield m
                except QuitException:
                    pass
            container.subroutine(clearup())
        
def multiwaitif(references, container, expr, nextchange = False):
    keys = set(r.getkey() for r in references)
    matchers = tuple(DataObjectUpdateEvent.createMatcher(k) for k in keys)
    flag = nextchange
    def updateref():
        e = container.event
        k = e.key
        o = e.object
        for r in references:
            if r.getkey() == k:
                if r._ref is not o:
                    r._ref = o
    updated_keys = ()
    while True:
        updated_values = [ref for ref in references if ref.getkey() in updated_keys]
        r = not flag and expr(references, updated_values)
        flag = False
        if r:
            container.retvalue = ([ref for ref in references if ref.getkey() in updated_keys], r)
            break
        yield matchers
        transid = container.event.transactid
        updated_keys = keys.intersection(container.event.allkeys)
        last_key = [k for k in container.event.allkeys if k in keys][-1]
        transact_matcher = DataObjectUpdateEvent.createMatcher(None, transid, _ismatch = lambda x: x.key in updated_keys)
        while True:
            updateref()
            if container.event.key == last_key:
                break
            yield (transact_matcher,)

def updater(f):
    "Decorate a function with named arguments into updater for transact"
    @functools.wraps(f)
    def wrapped_updater(keys, values):
        result = f(*values)
        return (keys[:len(result)], result)
    return wrapped_updater

def list_updater(*args):
    """
    Decorate a function with named lists into updater for transact.
    
    :params \*args: parameter list sizes. -1 means all other items. None means a single item instead of a list.
                    only one -1 is allowed.
    """
    neg_index = [i for v,i in izip(args, itertools.count()) if v is not None and v < 0]
    if len(neg_index) > 1:
        raise ValueError("Cannot use negative values more than once")
    if not neg_index:
        slice_list = []
        size = 0
        for arg in args:
            if arg is None:
                slice_list.append(size)
                size += 1
            else:
                slice_list.append(slice(size, size + arg))
                size += arg
    else:
        sep = neg_index[0]
        slice_list = []
        size = 0
        for arg in args[:sep]:
            if arg is None:
                slice_list.append(size)
                size += 1
            else:
                slice_list.append(slice(size, size + arg))
                size += arg
        rslice_list = []
        rsize = 0
        for arg in args[:sep:-1]:
            if arg is None:
                rslice_list.append(-1-rsize)
                rsize += 1
            else:
                rslice_list.append(slice(None if not rsize else -rsize, -(rsize + arg)))
                rsize += arg
        slice_list.append(slice(size, rsize))
        slice_list.extend(reversed(rslice_list))
    def inner_wrapper(f):
        @functools.wraps(f)
        def wrapped_updater(keys, values):
            result = f(*[values[s] for s in slice_list])
            return (keys[:len(result)], result)
        return wrapped_updater
    return inner_wrapper

class AlreadyExistsException(Exception):
    pass

def create_new(cls, oldvalue, *args):
    "Raise  if the old value already exists"
    if oldvalue is not None:
        raise AlreadyExistsException('%r already exists' % (oldvalue,))
    return cls.create_instance(*args)

def create_from_key(cls, oldvalue, key):
    "Raise  if the old value already exists"
    if oldvalue is not None:
        raise AlreadyExistsException('%r already exists' % (oldvalue,))
    return cls.create_from_key(key)

def set_new(oldvalue, newvalue):
    if oldvalue is not None:
        raise AlreadyExistsException('%r already exists' % (oldvalue,))
    return newvalue.clone_instance()

def dump(obj, attributes = True, _refset = None):
    "Show full value of a data object"
    if _refset is None:
        _refset = set()
    if obj is None:
        return None
    elif isinstance(obj, DataObject):        
        if id(obj) in _refset:
            attributes = False
        else:
            _refset.add(id(obj))
        cls = type(obj)
        clsname = getattr(cls, '__module__', '<unknown>') + '.' + getattr(cls, '__name__', '<unknown>')
        baseresult = {'_type': clsname, '_key': obj.getkey()}
        if not attributes:
            return baseresult
        else:
            baseresult.update((k,dump(v, attributes, _refset)) for k,v in vars(obj).items() if k[:1] != '_')
            _refset.remove(id(obj))
        return baseresult
    elif isinstance(obj, ReferenceObject):
        if obj._ref is not None:
            return dump(obj._ref, attributes, _refset)
        else:
            return {'_ref':obj.getkey()}
    elif isinstance(obj, WeakReferenceObject):
        return {'_weakref':obj.getkey()}
    elif isinstance(obj, DataObjectSet):
        return dump(list(obj.dataset()))
    elif isinstance(obj, dict):
        return dict((k, dump(v, attributes, _refset)) for k,v in obj.items())
    elif isinstance(obj, list) or isinstance(obj, tuple) or isinstance(obj, set):
        return [dump(v, attributes, _refset) for v in obj]
    else:
        return obj



'''
Created on 2015/10/19

:author: hubo
'''
from vlcp.server.module import Module, api
from vlcp.event import TcpServer
from vlcp.event.runnable import RoutineContainer
from vlcp.event.connection import Client

class TcpServerBase(Module):
    '''
    Generic tcp server on specified URLs, vHosts are supported.
    '''
    # Default server URL.
    # 
    # If there are multiple endpoints to listen, use .urls configuration instead of .url,
    # it should be a list of endpoint URLs.
    # 
    # The URL (and also any "connection URL" in other part of VLCP) should be like::
    # 
    #    <protocol>://[<address>[:port]]/
    # 
    # For server socket (the listening part), the protocol should be tcp(or ltcp), ssl(or lssl)
    # The address should be the binding address of the server.
    # 
    # For client sockets (the connecting part), the protocol should be tcp, udp, ssl, ptcp, pudp, pssl
    # 
    # The ptcp, pudp, pssl protocols are passive connections: they start listen and accept only one socket.
    # The address (if specified) should be the IP address of the remote endpoint.
    # 
    # The tcp, udp, ssl protocols are normal connections which begins by connecting to the specified address.
    # 
    # If .urls and .url are both specified, they are both used.
    #
    # For SSL connections, use .key, .certificate, .ca_certs configurations to specify private key, public key,
    # CA files for the SSL connection. If .ca_certs is not specified, the SSL certificate is not verified which
    # may introduce security holes.
    #
    # the .protocol configuration node may be used to override the global protocol configurations, e.g. use::
    # 
    #     module.redisdb.protocol.connect_timeout = 20
    # 
    # to override the protocol.redis.connect_timeout configuration
    #
    # Any TCPServer module can use .vhost node to create multiple servers for different uses.
    # For example, use::
    # 
    #     module.httpserver.url='tcp://localhost:80/'
    #     module.httpserver.vhost.api.url='tcp://localhost:8080/'
    #     module.httpserver.vhost.api.protocol.fastfail=True
    # 
    # to create two different HTTP servers, potentially with different configurations. Some modules can be
    # configured to bind to specified vhosts.
    #
    # All configurations for the default vhost ('') can be used on a vhost, like .urls and .protocol and even
    # .vhost. Vhosts inherits the settings from the parent node (the default server or another vhost) except URLs.
    _default_url = 'tcp:///'
    # If True, the incoming connections are automatically managed, and can be queried with *getconnections* API.
    # Also when the module is unloaded, the incoming connections are closed.
    _default_connmanage = False
    # By default the .url and .urls are recognized as server URLs, so it creates listening sockets.
    # use client=True to make this module recognize the URLs as client URLs, so it creates client connections
    # (either normal or passive)
    _default_client = False
    service = True
    def _createprotocol(self, config):
        return self._protocolclass()
    def _createServers(self, config, vhostname, defaultconfig = {}, key = None, certificate = None, ca_certs = None, exists = {}, client = False):
        urls = list(getattr(config, 'urls', []))
        if hasattr(config, 'url') and config.url and config.url.strip():
            urls.append(config.url.strip())
        settings = dict(defaultconfig.items())
        if hasattr(config, 'protocol'):
            settings.update(config.protocol.items())
        key = getattr(config, 'key', key)
        certificate = getattr(config, 'certificate', certificate)
        ca_certs = getattr(config, 'ca_certs', ca_certs)
        client = getattr(config, 'client', client)
        if urls:
            defaultProtocol = self._createprotocol(config)
            if self.connmanage:
                # Patch init() and final()
                def patch(prococol):
                    orig_init = prococol.init
                    def init(conn):
                        for m in orig_init(conn):
                            yield m
                        self.managed_connections.add(conn)
                    prococol.init = init
                
                    orig_final = prococol.final
                    def final(conn):
                        try:
                            self.managed_connections.remove(conn)
                        except Exception:
                            pass
                        for m in orig_final(conn):
                            yield m
                    prococol.final = final
                patch(defaultProtocol)
            defaultProtocol.vhost = vhostname
            # Copy extra configurations to protocol
            for k,v in settings:
                setattr(defaultProtocol, k, v)
            for url in urls:
                if (vhostname, url) in exists:
                    exists.remove((vhostname, url))
                else:
                    if client:
                        self.connections.append(self._client_class(config, defaultProtocol, vhostname)(url, defaultProtocol, self.scheduler,
                                                       key, certificate, ca_certs, getattr(config, 'bindaddress', None)))
                    else:
                        self.connections.append(self._server_class(config, defaultProtocol, vhostname)(url, defaultProtocol, self.scheduler,
                                                      key, certificate, ca_certs))
        if hasattr(config, 'vhost'):
            for k,v in config.vhost.items():
                self._createServers(v, k, settings, key, certificate, ca_certs, exists, client)
    def _client_class(self, config, protocol, vhost):
        return Client
    def _server_class(self, config, protocol, vhost):
        return TcpServer
    def __init__(self, server, protocolclass):
        Module.__init__(self, server)
        self._protocolclass = protocolclass
        self.apiroutine = RoutineContainer(self.scheduler)
        self.managed_connections = set()
        self._createServers(self, '')
        self.createAPI(api(self.getservers),
                       api(self.stoplisten, self.apiroutine),
                       api(self.startlisten, self.apiroutine),
                       api(self.updateconfig, self.apiroutine),
                       api(self.getconnections))
    def unload(self, container, force=False):
        if self.connmanage:
            self.connections.extend(self.managed_connections)
            self.managed_connections.clear()
        for m in Module.unload(self, container, force=force):
            yield m
    def getservers(self, vhost = None):
        '''
        Return current servers
        
        :param vhost: return only servers of vhost if specified. '' to return only default servers.
                      None for all servers.
        '''
        if vhost is not None:
            return [s for s in self.connections if s.protocol.vhost == vhost]
        else:
            return list(self.connections)
    def stoplisten(self, vhost = None):
        '''
        Stop listen on current servers
        
        :param vhost: return only servers of vhost if specified. '' to return only default servers.
                      None for all servers.
        '''
        servers = self.getservers(vhost)
        for s in servers:
            for m in s.stoplisten():
                yield m
        self.apiroutine.retvalue = len(servers)
    def startlisten(self, vhost = None):
        '''
        Start listen on current servers
        
        :param vhost: return only servers of vhost if specified. '' to return only default servers.
                      None for all servers.
        '''
        servers = self.getservers(vhost)
        for s in servers:
            for m in s.startlisten():
                yield m
        self.apiroutine.retvalue = len(servers)
    def updateconfig(self):
        "Reload configurations, remove non-exist servers, add new servers, and leave others unchanged"
        exists = {}
        for s in self.connections:
            exists[(s.protocol.vhost, s.rawurl)] = s
        self._createServers(self, '', exists = exists)
        for _,v in exists.items():
            for m in v.shutdown():
                yield m
            self.connections.remove(v)
        self.apiroutine.retvalue = None
    def getconnections(self, vhost = None):
        "Return accepted connections, optionally filtered by vhost"
        if vhost is None:
            return list(self.managed_connections)
        else:
            return [c for c in self.managed_connections if c.protocol.vhost == vhost]



'''
Created on 2015/11/26

:author: hubo
'''

from namedstruct import *
from namedstruct import sizefromlen, packvalue

def _pack_xlen(ns):
    ns.xlen = ns._realsize() - 2

extra = nstruct((uint16_le, 'xlen'),
                name='extra',
                size=lambda x: x.xlen + 2,
                prepack=_pack_xlen,
                padding=1,
                endian='<'
                )

flg = enum('flg', globals(), uint8_le, True,
           FTEXT = 1<<0,
           FHCRC = 1<<1,
           FEXTRA = 1<<2,
           FNAME = 1<<3,
           FCOMMENT = 1<<4
           )

_flags = [('extra', FEXTRA),
          ('fname', FNAME),
          ('fcomment', FCOMMENT),
          ('fhcrc', FHCRC)]

def _pack_flg(s):
    for name,f in _flags:
        if hasattr(s, name):
            s.flg = s.flg | f
        else:
            s.flg = s.flg & (~f)


def _init_header(s):
    s.id1 = 0x1f
    s.id2 = 0x8b
    s.cm = 8
    s.xfl = 2
    s.os = 255

header = nstruct(
                 (uint8_le, 'id1'),
                 (uint8_le, 'id2'),
                 (uint8_le, 'cm'),
                 (flg, 'flg'),
                 (uint32_le, 'mtime'),
                 (uint8_le, 'xfl'),
                 (uint8_le, 'os'),
                 (optional(extra, 'extra', lambda x: x.flg & FEXTRA),),
                 (optional(cstr, 'fname', lambda x: x.flg & FNAME),),
                 (optional(cstr, 'fcomment', lambda x: x.flg & FCOMMENT),),
                 (optional(uint16_le, 'fhcrc', lambda x: x.flg & FHCRC),),
                 name = 'header',
                 endian = '<',
                 padding = 1,
                 init = _init_header,
                 prepack=_pack_flg
                 )

tail = nstruct(
               (uint32_le, 'crc32'),
               (uint32_le, 'isize'),
               endian = '<',
               name = 'tail',
               padding = 1
               )



'''
Created on 2016/7/26

:author: think
'''
from vlcp.utils.ethernet import ip4_addr

def parse_ip4_network( network ):

    ip,f,prefix = network.rpartition('/')
    if not f or not prefix:
        raise ValueError('invalid cidr ' + network)
    if not 0 <= int(prefix) <= 32:
        raise ValueError("invalid prefix " + prefix)

    netmask = get_netmask(prefix)

    value = ip4_addr(ip)
    return value & netmask,int(prefix)

def get_netmask(prefix):
    return (0xffffffff) >> (32 - int(prefix)) << (32 - int(prefix))

def get_network(ip, prefix):
    return ip & get_netmask(prefix)

def get_broadcast(network, prefix):
    return network | ((1 << (32 - prefix)) - 1)

def parse_ip4_address(address):
    return ip4_addr(address)

def ip_in_network(ip,network,prefix):
    shift = 32 - prefix
    return (ip >> shift) == (network >> shift)

def network_first(network,prefix):
    return network + 1

def network_last(network,prefix):
    hostmask = (1 << (32 - prefix)) - 1
    # calc cidr last avaliable ip , so inc 1
    return (network | hostmask) - 1

def format_network_cidr(cidr):
    ip,f,prefix = cidr.rpartition('/')

    check_ip_address(ip)

    if f and prefix:
        return cidr
    else:
        return ip + "/32"

def check_ip_address(ipaddress):
    try:
        ip4_addr(ipaddress)
    except Exception:
        raise ValueError(" ip address illegal " + ipaddress)

    return ipaddress


'''
Created on 2015/12/29

:author: hubo
'''
from __future__ import print_function
from vlcp.utils.connector import async_processor, async_to_async, Connector,\
    generator_to_async
from vlcp.event.event import withIndices, Event
from vlcp.config import defaultconfig
from vlcp.server.module import Module, api, callAPI
import functools
import threading
import signal
from vlcp.event.runnable import RoutineContainer
from vlcp.event.runnable import RoutineException
import pdb
import code
from vlcp.config.config import manager
from vlcp.protocol.protocol import Protocol
from vlcp.event.connection import Client
import os
import socket
import re
from vlcp.event.core import InterruptedBySignalException
try:
    from Queue import Queue, PriorityQueue
except:
    from queue import Queue, PriorityQueue
import traceback
import sys
try:
    import thread
except:
    import _thread as thread

def console_help():
    print(Console._full_help)

def restore_console():
    if not hasattr(Console, '_instance') or not Console._instance:
        raise ValueError('Console is not loaded')
    Console._instance.restore_console()

@withIndices('type')
class ConsoleEvent(Event):
    canignore = False

@withIndices()
class ConsoleServiceCall(Event):
    pass

@withIndices('waiter')
class ConsoleServiceCancel(Event):
    pass

@withIndices('socket')
class SocketInjectDone(Event):
    pass

@withIndices()
class InterruptPoller(Event):
    pass

class Waiter(object):
    def __init__(self):
        self.event = threading.Event()
        self.event.clear()
        self.exception = None
        self.result = None
    def wait(self, timeout = None):
        self.event.wait(timeout)
        if self.exception:
            raise self.exception
        else:
            return self.result
    def raise_exception(self, exc):
        self.exception = exc
        self.event.set()
    def send_result(self, val):
        self.result = val
        self.event.set()

@defaultconfig
class Console(Module):
    '''
    VLCP debugging console.
    
    Besides the normal functions of Python interactive console,
    Following variables are provided for debugging purpose:
    
       server, manager, container
        
    Following functions can be used to control VLCP running:
    
       callapi, capture, sendevent, subroutine, execute, breakpoint, syscall,
       resume, debug, restore_console, console_help
       
    For details call console_help()
    '''
    _full_help = '''
VLCP debugging console.
Besides the normal functions of python interactive console,
following variables are provided for debugging purpose:
server - current running VLCP server
manager - current configuration manager
container - internal used routine container

Following functions can be used to control VLCP running:

callapi(modulename, functionname, **kwargs)
 - Call module API modulename/functionname with kwargs, return result

capture(matchers, blocking = False, breakpoint = False, captureonce = False, callback = None)
 - Capture events matched with specified matchers and print the event. Other parameters:
   - blocking: if True, wait until the events are captured
   - breakpoint: if True, suspend the event loop and wait for resume()
   - captureonce: if True, remove the matchers on first capture
   - callback: func(event, matcher) called on every capture if specified
   
sendevent(event, emerge = False)
 - Send specified event to scheduler. if merge = True, send immediately without block

subroutine(routine)
 - create a new routine in container.

execute(routine)
 - execute the routine in container, and return container.retvalue
 
breakpoint()
 - stop running and wait for resume().
 
syscall(syscall_func)
 - execute syscall_func in syscall context
 
resume()
 - resume from breakpoint
 
debug()
 - resume from breakpoint with pdb.set_trace() to enter pdb debugging. Suspend the interactive console
   to work with pdb.

restore_console()
 - Prepare to continue in pdb and resume the console. Type in pdb: 
     clear
     import vlcp.service.debugging.console
     vlcp.service.debugging.console.restore_console()
     continue

console_help()
 - show this help
'''
    service = False
    # Directly start VLCP in the console mode. By default, the console module creates a
    # telnet server and wait for a connection. The console can be used in the telnet session.
    # With startinconsole = True, the module uses stdin/stdout to create the console.
    _default_startinconsole = False
    # Default telnet connection URL, this is a passive connection on port 9923, so use::
    # 
    #     telnet localhost 9923
    # 
    # to connect to the console.
    _default_telnetconsole = 'ptcp://localhost:9923/'
    # If SSL is configured (with pssl://...), specify the private key file
    _default_key = None
    # If SSL is configured, specify the certificate file
    _default_certificate = None
    # If SSL is configured, specify the CA file
    _default_ca_certs = None
    def _service_routine(self):
        self.apiroutine.subroutine(self._intercept_main())
        csc = ConsoleServiceCall.createMatcher()
        while True:
            yield (csc,)
            self.apiroutine.subroutine(self.apiroutine.event.routine, True)
    def _service_call_routine(self, waiter, call):
        try:
            for m in self.apiroutine.withException(call, ConsoleServiceCancel.createMatcher(waiter)):
                yield  m
        except RoutineException:
            pass
        except Exception as exc:
            waiter.raise_exception(exc)
        else:
            waiter.send_result(self.apiroutine.retvalue)
    def _intercept_main(self):
        cr = self.apiroutine.currentroutine
        self.sendEventQueue = Queue()
        _console_connect_event = threading.Event()
        _console_connect_event.clear()
        for m in self.apiroutine.waitForSend(ConsoleEvent('initproxy')):
            yield m
        if not self.startinconsole:
            p = Protocol()
            p.persist = True
            p.createqueue = False
            def init(connection):
                sock = connection.socket
                self.telnet_socket = sock
                self.scheduler.unregisterPolling(connection.socket)
                connection.socket = None
                connection.connected = False
                _console_connect_event.set()
                yield (SocketInjectDone.createMatcher(sock),)
            p.init = init
            p.reconnect_init = init
            Client(self.telnetconsole, p, self.scheduler, self.key, self.certificate, self.ca_certs).start()
        def syscall_threaded_main(scheduler, processor):
            # Detach self
            scheduler.unregisterall(cr)
            self._threaded_main_quit = False
            def threaded_main():
                try:
                    scheduler.main(False, False)
                finally:
                    self._threaded_main_quit = True
                    _console_connect_event.set()
            t = threading.Thread(target=threaded_main)
            t.daemon = True
            t.start()
            try:
                if self.startinconsole:
                    self._interactive()
                else:
                    while not self._threaded_main_quit:
                        try:
                            while not _console_connect_event.is_set():
                                # There is a bug in Python 2.x that wait without timeout cannot be
                                # interrupted by signal
                                _console_connect_event.wait(3600)
                            if self._threaded_main_quit:
                                break
                        except InterruptedBySignalException:
                            # This signal should interrupt the poller, but poller is not in the main thread
                            # Send an event through the proxy will do the trick
                            self.sendEventQueue.put((InterruptPoller(),))
                            continue
                        pstdin_r, pstdin_w = os.pipe()
                        pstdout_r, pstdout_w = os.pipe()
                        orig_stdin = sys.stdin
                        orig_stdout = sys.stdout
                        orig_stderr = sys.stderr
                        try:
                            pstdin = os.fdopen(pstdin_r, 'rU', 0)
                            pstdout = os.fdopen(pstdout_w, 'w', 0)
                            sys.stdin = pstdin
                            sys.stdout = pstdout
                            sys.stderr = pstdout
                            sock = self.telnet_socket
                            sock.setblocking(True)
                            self.telnet_socket = None
                            _console_connect_event.clear()
                            t = threading.Thread(target=self._telnet_server, args=(pstdin_w, pstdout_r, sock, orig_stdout))
                            t.daemon = True
                            t.start()
                            try:
                                self._interactive()
                            except SystemExit:
                                pass
                            if not t.is_alive():
                                break
                            self.sendEventQueue.put((SocketInjectDone(sock),))
                        finally:
                            try:
                                sock.shutdown(socket.SHUT_RDWR)
                            except:
                                pass
                            try:
                                pstdin.close()
                            except:
                                pass
                            try:
                                pstdout.close()
                            except:
                                pass
                            sys.stdin = orig_stdin
                            sys.stdout = orig_stdout
                            sys.stderr = orig_stderr
            except SystemExit:
                pass
            finally:
                self.sendEventQueue.put(None)
                scheduler.quit()
                if self.startinconsole:
                    print('Wait for scheduler end, this may take some time...')
                t.join()
        for m in self.apiroutine.syscall(syscall_threaded_main, True):
            yield m
    def _telnet_server_writer(self, queue, sock):
        lastseq = -1
        while True:
            t, seq, val = queue.get()
            if t < 0:
                break
            if t != 2 or seq >= lastseq:
                try:
                    sock.sendall(val)
                except:
                    break
            if t == 0:
                lastseq = seq
    def _telnet_server_writer2(self, pstdout_r, queue, lock, orig_stdout):
        while True:
            data = os.read(pstdout_r, 1024)
            if data == '':
                os.close(pstdout_r)
                break
            data, _ = re.subn(br'\r?\n', b'\r\n', data)
            lock.acquire()
            try:
                self._telnet_seq += 1
                seq = self._telnet_seq
            finally:
                lock.release()
            queue.put((2, seq, data))
    def _telnet_server(self, pstdin_w, pstdout_r, sock, orig_stdout):
        queue = PriorityQueue()
        inputbuffer = b''
        self._telnet_seq = 0
        try:
            t = threading.Thread(target=self._telnet_server_writer, args=(queue, sock))
            t.daemon = True
            t.start()
            lock = threading.Lock()
            def writeall(data):
                start = 0
                while start < len(data):
                    size = os.write(pstdin_w, data[start:])
                    start += size
            def sendcontrol(t, data):
                lock.acquire()
                try:
                    self._telnet_seq += 1
                    seq = self._telnet_seq
                finally:
                    lock.release()
                queue.put((t, seq, data))
            t2 = threading.Thread(target=self._telnet_server_writer2, args=(pstdout_r, queue, lock, orig_stdout))
            t2.daemon = True
            t2.start()
            escaping = False
            option = None
            while True:
                newdata = sock.recv(1024)
                if newdata == b'':
                    break
                for i in range(0, len(newdata)):
                    c = newdata[i:i+1]
                    if escaping:
                        if option == b'\xfd' and c == b'\x06':
                            sendcontrol(1, b'\xff\xfb\x06')
                            option = None
                            escaping = False
                        elif option == b'\xfd' or option == b'\xfe':
                            sendcontrol(1, b'\xff\xfc' + c)
                            option = None
                            escaping = False
                        elif option == b'\xfb' or option == b'\xfc':
                            sendcontrol(1, b'\xff\xfe' + c)
                            option = None
                            escaping = False
                        elif c in (b'\xfb', b'\xfc', b'\xfd', b'\xfe'):
                            option = c
                        else:
                            option = None
                            if c == b'\xf3' or c == b'\xf4':
                                thread.interrupt_main()
                            escaping = False
                    else:
                        if c == b'\x03':
                            thread.interrupt_main()
                        elif c == b'\x08':
                            inputbuffer = inputbuffer[:-1]
                        elif c == b'\x00':
                            inputbuffer += b'\n'
                            writeall(inputbuffer)
                            inputbuffer = b''
                        elif c == b'\r' or c == b'\n':
                            inputbuffer += c
                            writeall(inputbuffer)
                            inputbuffer = b''
                        elif c == b'\xff':
                            escaping = True
                        else:
                            inputbuffer += c
        except OSError:
            pass
        except IOError:
            pass
        finally:
            try:
                os.close(pstdin_w)
            except:
                pass
            queue.put((-1, -1, -1))
    def _interactive(self):
        lsignal = signal.signal(signal.SIGINT, signal.default_int_handler)
        try:
            _breakpoint_event = threading.Event()
            _current_thread = threading.current_thread().ident
            _enter_pdb = [False]
            def _async_run(call):
                self.sendEventQueue.put((ConsoleServiceCall(routine = call),))
            def _async(func):
                @functools.wraps(func)
                def f(*args, **kwargs):
                    _async_run(func(*args, **kwargs))
                return f
            def _service_call_customized(factory):
                waiter = Waiter()
                self.sendEventQueue.put((ConsoleServiceCall(routine=factory(waiter)),))
                try:
                    return waiter.wait()
                except:
                    self.sendEventQueue.put((ConsoleServiceCancel(waiter),))
                    raise
            def execute(call):
                return _service_call_customized(lambda waiter: self._service_call_routine(waiter, call))
            def _service(func):
                @functools.wraps(func)
                def f(*args, **kwargs):
                    return execute(func(*args, **kwargs))
                return f
            @_service
            def callapi(modulename, functionname, **kwargs):
                return callAPI(self.apiroutine, modulename, functionname, kwargs)
            @_service
            def sendevent(event, emerge = False):
                if emerge:
                    self.apiroutine.scheduler.emergesend(event)
                else:
                    for m in self.apiroutine.waitForSend(event):
                        yield m
                self.apiroutine.retvalue = None
            @_service
            def subroutine(routine):
                self.apiroutine.retvalue = self.apiroutine.subroutine(routine)
                if False:
                    yield
            @_service
            def syscall(syscall_func):
                for m in self.apiroutine.syscall(syscall_func):
                    yield m
            def breakpoint():
                in_thread = threading.current_thread().ident
                if in_thread == _current_thread:
                    _breakpoint()
                else:
                    print('Enter VLCP debugging breakpoint:')
                    traceback.print_stack()
                    print('Call resume() to continue the event loop, or debug() to enter pdb')
                    _breakpoint_event.clear()
                    _breakpoint_event.wait()
                    if _enter_pdb[0]:
                        pdb.set_trace()
                    else:
                        print('Resume from breakpoint.')
            @_async
            def _breakpoint():
                breakpoint()
                if False:
                    yield
            def resume():
                _enter_pdb[0] = False
                _breakpoint_event.set()
            @_async
            def restore_console():
                self._restore_console_event.set()
                if False:
                    yield
            self.restore_console = restore_console
            def debug():
                _enter_pdb[0] = True
                self._restore_console_event.clear()
                _breakpoint_event.set()
                # Switch to event loop thread, suspend the main thread, wait for restore_console
                self._restore_console_event.wait()
            _capture_breakpoint = breakpoint
            def capture(matchers, blocking = False, breakpoint = False, captureonce = False, callback = None):
                def _capture_service(waiter):
                    if blocking:
                        csm = ConsoleServiceCancel.createMatcher(waiter)
                    else:
                        waiter.send_result(self.apiroutine.currentroutine)
                    firsttime = True
                    while firsttime or not captureonce:
                        if blocking:
                            yield tuple(matchers) + (csm,)
                        else:
                            yield matchers
                        if blocking and self.apiroutine.matcher is csm:
                            # Cancelled
                            return
                        print('Event Captured: Capture %r with %r' % (self.apiroutine.event, self.apiroutine.matcher))
                        if firsttime and blocking:
                            waiter.send_result((self.apiroutine.event, self.apiroutine.matcher, self.apiroutine.currentroutine))
                        firsttime = False
                        if callback:
                            try:
                                callback(self.apiroutine.event, self.apiroutine.matcher)
                            except:
                                print('Exception while running callback:')
                                traceback.print_exc()
                        if breakpoint:
                            _capture_breakpoint()
                return _service_call_customized(_capture_service)
            code.interact(self.__doc__ + '\n' + 'Python ' + str(sys.version) + ' on ' + str(sys.platform),
                          None,
                          {'server':self.server,'manager':manager, 'container':self.apiroutine,
                                     'callapi':callapi, 'capture':capture, 'sendevent':sendevent,
                                     'subroutine':subroutine, 'breakpoint':breakpoint, 'syscall':syscall,
                                     'resume':resume, 'debug':debug, 'restore_console':restore_console,
                                     'console_help':console_help,'execute':execute})
        finally:
            signal.signal(signal.SIGINT, lsignal)
    def __init__(self, server):
        '''
        Constructor
        '''
        Module.__init__(self, server)
        self._ce_matcher = ConsoleEvent.createMatcher()
        self.apiroutine = RoutineContainer(self.scheduler)
        self.apiroutine.main = self._service_routine
        self._restore_console_event = threading.Event()
        @generator_to_async(True, False)
        def proxy(event, matcher):
            while True:
                events = self.sendEventQueue.get()
                if events is None:
                    break
                yield events
        @async_to_async(True, False)
        @async_processor
        def processor(event, matcher, queueout):
            if event.type == 'initproxy':
                proxy(event, matcher, queueout)
        self.connector = Connector(processor, (self._ce_matcher,), self.scheduler, False)
        self.routines.append(self.apiroutine)
        self.routines.append(self.connector)

if __name__ == '__main__':
    from vlcp.server import main
    import sys
    manager['module.console.startinconsole'] = True
    modules = list(sys.argv[1:]) + ['__main__.Console']
    main(None, modules)



'''
Created on 2015/6/19

:author: hubo
'''
from __future__ import print_function, absolute_import, division 
from .runnable import RoutineContainer, RoutineException
from .core import PollEvent, POLLING_ERR, POLLING_IN, POLLING_HUP
from .event import Event, withIndices
from vlcp.utils import ContextAdapter
from ctypes import create_string_buffer, c_char, Array as _Array, memmove as _memmove
import socket
import errno
import sys
import ssl
import multiprocessing
import threading
import logging
import itertools
import signal
from vlcp.event.core import POLLING_OUT
import traceback
from contextlib import closing
import os

if sys.version_info[0] >= 3:
    from urllib.parse import urlsplit
    from queue import Full, Queue, Empty
else:
    from urlparse import urlsplit
    from Queue import Full, Queue, Empty

try:
    _memoryview = memoryview
except:
    class memoryview(object):              # Fake a memoryview interface
        def __init__(self, buf):
            self._buffer = buf
        def __getitem__(self, s):
            if isinstance(s, slice) and isinstance(self._buffer, _Array):
                start, stop, step = s.indices(len(self._buffer))
                if step != 1:
                    raise ValueError('memoryview does not support slicing with step')
                return (c_char * (stop - start)).from_buffer(self._buffer, start)
            else:
                return self._buffer[s]
        def __setitem__(self, s, value):
            if isinstance(s, slice) and isinstance(self._buffer, _Array):
                start, stop, step = s.indices(len(self._buffer))
                if step != 1:
                    raise ValueError('memoryview does not support slicing with step')
                if len(value) != stop - start:
                    raise ValueError('assignment length not match')
                _memmove((c_char * (stop - start)).from_buffer(self._buffer, start), value, stop - start)
            else:
                self._buffer[s] = value
        def __len__(self):
            return len(self._buffer)
        def tobytes(self):
            return self._buffer[:]

if sys.version_info[0] >= 3:
    _create_buffer = bytearray
    def _extend_buffer(source, size):
        _new_buffer = _create_buffer(size)
        _new_buffer[0:len(source)] = source
        return _new_buffer
    def _buffer(data, view, start, length):
        return view[start:start+length]
else:
    _create_buffer = create_string_buffer
    def _extend_buffer(source, size):
        return create_string_buffer(source.raw, size)
    # re does not accept memoryview
    def _buffer(data, view, start, length):
        return buffer(data, start, length)

@withIndices('connection', 'type', 'force', 'connmark')
class ConnectionControlEvent(Event):
    SHUTDOWN = 'shutdown'
    RECONNECT = 'reconnect'
    RESET = 'reset'
    STOPLISTEN = 'stoplisten'
    STARTLISTEN = 'startlisten'
    HANGUP = 'hangup'

@withIndices('connection', 'connmark')
class ConnectionWriteEvent(Event):
    """
    Event used to send data to a connection
    """
    canignore = False
    def canignorenow(self):
        return not self.connection.connected or self.connection.connmark != self.connmark

class ConnectionResetException(Exception):
    pass

class Connection(RoutineContainer):
    '''
    A connection on a socket
    '''
    logger = logging.getLogger(__name__ +'.Connection')
    def __init__(self, protocol, sockobj = None, scheduler = None):
        '''
        Constructor
        '''
        RoutineContainer.__init__(self, scheduler)
        self.connected = False
        self.protocol = protocol
        self.socket = sockobj
        if self.socket is not None:
            self.socket.setblocking(False)
        self.persist = getattr(protocol, 'persist', False)
        self.need_reconnect = self.persist
        self.keepalivetime = getattr(protocol, 'keepalivetime', None)
        self.logContext = {'connection': self, 'protocol':protocol, 'socket':None if self.socket is None else self.socket.fileno()}
        self.logger = ContextAdapter(Connection.logger, {'context':self.logContext}) 
        # Counters
        self.totalrecv = 0
        self.totalsend = 0
        self.connrecv = 0
        self.connsend = 0
        self.daemon = False
    def attach(self, sockobj):
        self.socket = sockobj
        self.logContext['socket'] = None if self.socket is None else self.socket.fileno()
        self.logger = ContextAdapter(Connection.logger, {'context': self.logContext})
        if self.socket is not None:
            self.socket.setblocking(False)
    def setdaemon(self, daemon):
        if self.daemon != daemon:
            if self.socket:
                self.scheduler.setPollingDaemon(self.socket, daemon)
            if hasattr(self, 'mainroutine'):
                self.scheduler.setDaemon(self.mainroutine, daemon, True)
            self.daemon = daemon
    def _read_main(self):
        try:
            canread_matcher = PollEvent.createMatcher(self.socket.fileno(), PollEvent.READ_READY)
            canwrite_matcher = PollEvent.createMatcher(self.socket.fileno(), PollEvent.WRITE_READY)
            self.readstop = False
            keepalivetime = self.keepalivetime
            buffersize = getattr(self.protocol, 'buffersize', 4096)
            buf = _create_buffer(buffersize)
            view = memoryview(buf)
            currPos = 0
            lastPos = 0
            exitLoop = False
            connection_control = ConnectionControlEvent.createMatcher(self, None, False, _ismatch = lambda x: x.connmark == self.connmark or x.connmark < 0)
            firstTime = True
            wantWrite = False
            eofExit = False
            while not exitLoop:
                if not firstTime:
                    if wantWrite:
                        yield (canwrite_matcher, connection_control)
                    else:
                        for m in self.waitWithTimeout(keepalivetime, canread_matcher, connection_control):
                            yield m
                        if self.timeout:
                            self.subroutine(self.protocol.keepalive(self))
                            continue
                    wantWrite = False
                    if self.matcher is connection_control:
                        if self.event.type == ConnectionControlEvent.RECONNECT:
                            self.need_reconnect = True
                        elif self.event.type == ConnectionControlEvent.SHUTDOWN:
                            self.need_reconnect = False
                        else:
                            self.need_reconnect = self.persist
                        exitLoop = True
                else:
                    firstTime = False
                parsed = True
                while True:
                    try:
                        currLen = self.socket.recv_into(view[currPos:])
                        if currLen == 0:
                            eofExit = True
                            exitLoop = True
                            break
                        else:
                            parsed = False
                    except ssl.SSLError as exc_ssl:
                        if exc_ssl.args[0] == ssl.SSL_ERROR_WANT_READ:
                            break
                        elif exc_ssl.args[0] == ssl.SSL_ERROR_WANT_WRITE:
                            wantWrite = True
                            break
                        else:
                            exitLoop = True
                            break
                    except socket.error as exc:
                        if exc.errno == errno.EAGAIN or exc.errno == errno.EWOULDBLOCK:
                            break
                        elif exc.errno == errno.EINTR:
                            continue
                        else:
                            exitLoop = True
                            break
                    self.totalrecv += currLen
                    self.connrecv += currLen
                    currPos = currLen + currPos
                    if currPos >= len(buf):
                        try:
                            (events, keep) = self.protocol.parse(self, _buffer(buf, view, 0, currPos), lastPos)
                        except:
                            # An exception in parse means serious protocol break, drop the connection
                            # Shutdown will prevent further data been sent in the socket and break the connection
                            self.socket.shutdown(socket.SHUT_RDWR)
                            raise
                        view[0:keep] = view[currPos-keep:currPos]
                        newPos = keep
                        parsed = True
                        for e in events:
                            for m in self.waitForSend(e):
                                while True:
                                    yield m + (connection_control,)
                                    if self.matcher is connection_control:
                                        if self.event.type == ConnectionControlEvent.RECONNECT:
                                            self.need_reconnect = True
                                        elif self.event.type == ConnectionControlEvent.SHUTDOWN:
                                            self.need_reconnect = False
                                        else:
                                            self.need_reconnect = self.persist
                                        exitLoop = True
                                    else:
                                        break
                        currPos = newPos
                        lastPos = newPos
                        if currPos >= len(buf):
                            buffer2 = _extend_buffer(buf, len(buf) * 2)
                            buf = buffer2
                            view = memoryview(buf)
                        for m in self.doEvents():
                            while True:
                                yield m + (connection_control,)
                                if self.matcher is connection_control:
                                    if self.event.type == ConnectionControlEvent.RECONNECT:
                                        self.need_reconnect = True
                                    elif self.event.type == ConnectionControlEvent.SHUTDOWN:
                                        self.need_reconnect = False
                                    else:
                                        self.need_reconnect = self.persist
                                    exitLoop = True
                                else:
                                    break
                if not parsed:
                    (events, keep) = self.protocol.parse(self, _buffer(buf, view, 0, currPos), lastPos)
                    view[0:keep] = view[currPos-keep:currPos]
                    newPos = keep
                    currPos = newPos
                    lastPos = newPos
                    parsed = True
                    for e in events:
                        for m in self.waitForSend(e):
                            yield m
                if eofExit:
                    # An extra parse for socket shutdown, can be determined by lastPos == len(data)
                    (events, keep) = self.protocol.parse(self, _buffer(buf, view, 0, currPos), lastPos)
                    for e in events:
                        for m in self.waitForSend(e):
                            yield m                    
        finally:
            self.readstop = True
            if self.writestop and self.connected:
                self.scheduler.emergesend(ConnectionControlEvent(self, ConnectionControlEvent.HANGUP, True, self.connmark))
    def _write_main(self):
        try:
            self.writestop = False
            canread_matcher = PollEvent.createMatcher(self.socket.fileno(), PollEvent.READ_READY)
            canwrite_matcher = PollEvent.createMatcher(self.socket.fileno(), PollEvent.WRITE_READY)
            connection_control = ConnectionControlEvent.createMatcher(self, None, False, _ismatch = lambda x: x.connmark == self.connmark or x.connmark < 0)
            write_matcher = ConnectionWriteEvent.createMatcher(self, self.connmark)
            queue_empty = None
            exitLoop = False
            while not exitLoop:
                if queue_empty is not None:
                    yield (write_matcher, queue_empty)
                else:
                    yield (write_matcher, connection_control)
                if self.matcher is connection_control:
                    if self.event.type == ConnectionControlEvent.RECONNECT:
                        self.need_reconnect = True
                    elif self.event.type == ConnectionControlEvent.SHUTDOWN:
                        self.need_reconnect = False
                    else:
                        self.need_reconnect = self.persist
                    if hasattr(self, 'queue'):
                        queue_empty = self.queue.waitForEmpty()
                        if queue_empty is None:
                            exitLoop = True
                    else:
                        exitLoop = True
                elif self.matcher is queue_empty:
                    queue_empty = self.queue.waitForEmpty()
                    if queue_empty is None:
                        exitLoop = True
                else:
                    self.event.canignore = True
                    msg, isEOF = self.protocol.serialize(self, self.event)
                    msg = memoryview(msg)
                    totalLen = len(msg)
                    currPos = 0
                    while currPos < totalLen:
                        wouldblock = False
                        wantRead = False
                        try:
                            currLen = self.socket.send(msg[currPos:])
                            currPos += currLen
                            self.totalsend += currLen
                            self.connsend += currLen
                        except ssl.SSLError as exc_ssl:
                            if exc_ssl.args[0] == ssl.SSL_ERROR_WANT_READ:
                                wantRead = True
                                wouldblock = True
                            elif exc_ssl.args[0] == ssl.SSL_ERROR_WANT_WRITE:
                                wantRead = False
                                wouldblock = True
                            else:
                                raise
                        except socket.error as exc:
                            if exc.errno == errno.EAGAIN or exc.errno == errno.EWOULDBLOCK:
                                wouldblock = True
                            elif exc.errno == errno.EPIPE or exc.errno == errno.ECONNRESET or exc.errno == errno.ECONNABORTED:
                                exitLoop = True
                                break
                            elif exc.errno == errno.EINTR:
                                continue
                            else:
                                raise
                        if wouldblock:
                            if wantRead:
                                yield (canread_matcher, connection_control)
                            else:
                                yield (canwrite_matcher, connection_control)
                            if self.matcher is connection_control:
                                if self.event.type == ConnectionControlEvent.RECONNECT:
                                    self.need_reconnect = True
                                elif self.event.type == ConnectionControlEvent.SHUTDOWN:
                                    self.need_reconnect = False
                                else:
                                    self.need_reconnect = self.persist
                                if hasattr(self, 'queue'):
                                    queue_empty = self.queue.waitForEmpty()
                                    if queue_empty is None:
                                        exitLoop = True
                                else:
                                    exitLoop = True
                    if isEOF:
                        try:
                            self.socket.shutdown(socket.SHUT_WR)
                        except socket.error:
                            # The socket may already be closed or reset
                            pass
                        exitLoop = True
        finally:
            self.writestop = True
            if self.readstop and self.connected:
                self.scheduler.emergesend(ConnectionControlEvent(self, ConnectionControlEvent.HANGUP, True, self.connmark))
    def _reconnect(self):
        self.connected = False
        if False:
            yield
    def _close(self):
        if hasattr(self, 'readroutine'):
            self.readroutine.close()
        if hasattr(self, 'writeroutine'):
            self.writeroutine.close()
        if self.socket is not None:
            self.scheduler.unregisterPolling(self.socket, self.daemon)
            self.socket.close()
            self.socket = None
        self.connected = False
    def main(self):
        try:
            try:
                self.localaddr = self.socket.getsockname()
            except:
                pass
            try:
                self.remoteaddr = self.socket.getpeername()
            except:
                pass
            self.connmark = 0
            self.connected = True
            for m in self.protocol.init(self):
                yield m
            connection_control = ConnectionControlEvent.createMatcher(self, None, True, _ismatch = lambda x: x.connmark == self.connmark or x.connmark < 0)
            while self.connected or self.need_reconnect:
                if self.connected:
                    self.connrecv = 0
                    self.connsend = 0
                    self.writestop = False
                    self.readstop = False
                    self.subroutine(self._read_main(), True, 'readroutine', self.daemon)
                    self.subroutine(self._write_main(), True, 'writeroutine', self.daemon)
                    err_match = PollEvent.createMatcher(self.socket.fileno(), PollEvent.ERROR)
                    yield (err_match, connection_control)
                    if self.matcher is connection_control and self.event.type is not ConnectionControlEvent.HANGUP:
                        if self.event.type == ConnectionControlEvent.SHUTDOWN:
                            self.need_reconnect = False
                        elif self.event.type == ConnectionControlEvent.RECONNECT:
                            self.need_reconnect = True
                        else:
                            self.need_reconnect = self.persist
                        self.logger.debug('Connection shutdown request received')
                    else:
                        self.logger.debug('Connection is closed')
                    self.connected = False
                    self.readroutine.close()
                    self.writeroutine.close()
                    if self.matcher is err_match and (self.event.detail & POLLING_ERR):
                        for m in self.protocol.error(self):
                            yield m
                    else:
                        for m in self.protocol.closed(self):
                            yield m
                if self.need_reconnect:
                    self.logger.debug('Try reconnecting')
                    for m in self._reconnect():
                        yield m
                    if self.connected:
                        try:
                            self.localaddr = self.socket.getsockname()
                        except:
                            pass
                        try:
                            self.remoteaddr = self.socket.getpeername()
                        except:
                            pass
                        self.connmark += 1
                        for m in self.protocol.reconnect_init(self):
                            yield m
                    else:
                        break
        finally:
            need_close = self.connected
            self._close()
            self.subroutine(self._final(need_close), False)
    def _final(self, need_close = False):
        if need_close:
            self.logger.debug('System is quitting, close connection, call protocol.closed()')
            for m in self.protocol.closed(self):
                yield m
        for m in self.protocol.final(self):
            yield m
    def shutdown(self, force = False, connmark = -1):
        '''
        Can call without delegate
        '''
        if connmark is None:
            connmark = self.connmark
        for m in self.waitForSend(ConnectionControlEvent(self, ConnectionControlEvent.SHUTDOWN, force, connmark)):
            yield m
    def reconnect(self, force = True, connmark = None):
        '''
        Can call without delegate
        '''
        if connmark is None:
            connmark = self.connmark
        for m in self.waitForSend(ConnectionControlEvent(self, ConnectionControlEvent.RECONNECT, force, connmark)):
            yield m
    def reset(self, force = True, connmark = None):
        '''
        Can call without delegate
        '''
        if connmark is None:
            connmark = self.connmark
        for m in self.waitForSend(ConnectionControlEvent(self, ConnectionControlEvent.RESET, force, connmark)):
            yield m
    def write(self, event, ignoreException = True):
        '''
        Can call without delegate
        '''
        connmark = self.connmark
        if self.connected:
            for m in self.waitForSend(event):
                yield m
                if not self.connected or self.connmark != connmark:
                    if not ignoreException:
                        raise ConnectionResetException
                    else:
                        return
        else:
            if not ignoreException:
                raise ConnectionResetException
    def __repr__(self, *args, **kwargs):
        baserepr = RoutineContainer.__repr__(self, *args, **kwargs)
        return baserepr + '(%r -> %r)' % (getattr(self, 'localaddr', None), getattr(self, 'remoteaddr', None))
    

@withIndices('request')
class ResolveRequestEvent(Event):
    canignore = False

@withIndices('request')
class ResolveResponseEvent(Event):
    pass

class Client(Connection):
    '''
    A single connection to a specified target
    '''
    def __init__(self, url, protocol, scheduler = None, key = None, certificate = None, ca_certs = None, bindaddress = None):
        Connection.__init__(self, protocol, None, scheduler)
        self.rawurl = url
        self.url = urlsplit(url, 'tcp')
        if self.url.scheme == 'ptcp':
            self.udp = False
            self.passive = True
            self.ssl = False
            self.unix = False
        elif self.url.scheme == 'unix':
            self.udp = False
            self.passive = False
            self.ssl = False
            self.unix = True
        elif self.url.scheme == 'punix':
            self.udp = False
            self.passive = True
            self.ssl = False
            self.unix = True
        elif self.url.scheme == 'ssl':
            self.udp = False
            self.passive = False
            self.ssl = True
            self.unix = False
        elif self.url.scheme == 'pssl':
            self.udp = False
            self.passive = True
            self.ssl = True
            self.unix = False
        elif self.url.scheme == 'udp':
            self.udp = True
            self.passive = False
            self.ssl = False
            self.unix = False
        elif self.url.scheme == 'pudp':
            self.udp = True
            self.passive = True
            self.ssl = False
            self.unix = False
        elif self.url.scheme == 'dunix':
            self.udp = True
            self.passive = False
            self.ssl = False
            self.unix = True
        elif self.url.scheme == 'pdunix':
            self.udp = True
            self.passive = True
            self.ssl = False
            self.unix = True
        else:
            self.udp = False
            self.passive = False
            self.ssl = False
            self.unix = False
        if not self.unix and not self.passive and not self.url.hostname:
            raise ValueError('Target address is not specified in url: ' + url)
        if self.unix and not self.url.path:
            raise ValueError('Unix socket path is not specified in url: ' + url)
        if not self.unix:
            self.hostname = self.url.hostname
            if not self.url.port:
                if self.ssl:
                    self.port = getattr(self.protocol, 'ssldefaultport', self.protocol.defaultport)
                else:
                    self.port = self.protocol.defaultport
            else:
                self.port = self.url.port
        else:
            self.path = self.url.path
        self.key = key
        self.certificate = certificate
        self.ca_certs = ca_certs
        self.bindaddress = bindaddress
        self.connect_timeout = getattr(self.protocol, 'connect_timeout', 30)
        self.reconnect_timeseq = getattr(self.protocol, 'reconnect_timeseq', self.defaultTimeSeq)
        self.reuseport = getattr(self.protocol, 'reuseport', False)
        self.nodelay = getattr(self.protocol, 'tcp_nodelay', False)
        self.logContext['url'] = self.rawurl
        if self.ssl:
            self.sslversion = getattr(self.protocol, 'sslversion', ssl.PROTOCOL_SSLv23)
    @staticmethod
    def defaultTimeSeq():
        nextSeq = 1
        while True:
            yield nextSeq                
            if nextSeq < 30:
                nextSeq = nextSeq * 2
            elif nextSeq < 270:
                nextSeq = nextSeq + 30
            else:
                nextSeq = 300
    def create_socket(self):
        if self.socket is not None:
            self.scheduler.unregisterPolling(self.socket)
            self.socket.close()
            self.socket = None            
        if not self.unix and self.hostname:
            request = (self.hostname, None if self.passive else self.port, socket.AF_UNSPEC, socket.SOCK_DGRAM if self.udp else socket.SOCK_STREAM, socket.IPPROTO_UDP if self.udp else socket.IPPROTO_TCP, socket.AI_ADDRCONFIG)
            # Resolve hostname
            for m in self.waitForSend(ResolveRequestEvent(request)):
                yield m
            for m in self.waitWithTimeout(self.connect_timeout, ResolveResponseEvent.createMatcher(request)):
                yield m
            if self.timeout:
                # Resolve is only allowed through asynchronous resolver
                #try:
                #    self.addrinfo = socket.getaddrinfo(self.hostname, self.port, socket.AF_UNSPEC, socket.SOCK_DGRAM if self.udp else socket.SOCK_STREAM, socket.IPPROTO_UDP if self.udp else socket.IPPROTO_TCP, socket.AI_ADDRCONFIG|socket.AI_NUMERICHOST)
                #except:
                raise IOError('Resolve hostname timeout: ' + self.hostname)
            else:
                if hasattr(self.event, 'error'):
                    raise IOError('Cannot resolve hostname: ' + self.hostname)
                self.addrinfo = self.event.response
        if self.passive:
            if self.unix:
                socket_listen = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM if self.udp else socket.SOCK_STREAM)
                socket_listen.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                if self.reuseport:
                    socket_listen.setsockopt(socket.SOL_SOCKET, 15, 1)                    
                self.logger.debug('Bind unix socket to %s', self.path)
                socket_listen.bind(self.path)
                self.valid_addresses = None
                family = socket.AF_UNIX
            elif self.hostname:
                if self.bindaddress is not None:
                    bind_protos = set(b[0] for b in self.bindaddress)
                else:
                    bind_protos = set((socket.AF_INET,socket.AF_INET6))
                addrinfo_protos = set(addr[0] for addr in self.addrinfo)
                if socket.AF_INET in bind_protos and socket.AF_INET in addrinfo_protos:
                    family = socket.AF_INET
                elif socket.AF_INET6 in bind_protos and socket.AF_INET6 in addrinfo_protos:
                    family = socket.AF_INET6
                else:
                    raise ValueError('Target address has different address family with local bind address')
                socket_listen = socket.socket(family, socket.SOCK_DGRAM if self.udp else socket.SOCK_STREAM)
                socket_listen.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                if not self.udp and self.nodelay:
                    socket_listen.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
                if self.reuseport:
                    socket_listen.setsockopt(socket.SOL_SOCKET, 15, 1)                    
                if self.bindaddress is not None:
                    for b in self.bindaddress:
                        if b[0] == family:
                            socket_listen.bind((b[1], self.port))
                            break
                else:
                    if family == socket.AF_INET:
                        socket_listen.bind(('0.0.0.0',self.port))
                    else:
                        socket_listen.bind(('::',self.port))
                self.valid_addresses = set(addr[4][0] for addr in self.addrinfo if addr[0] == family)
            else:
                if self.bindaddress is not None:
                    bind_protos = set(b[0] for b in self.bindaddress)
                else:
                    bind_protos = set((socket.AF_INET,socket.AF_INET6))
                if socket.AF_INET in bind_protos:
                    family = socket.AF_INET
                elif socket.AF_INET6 in bind_protos:
                    family = socket.AF_INET6
                else:
                    raise ValueError('Local bind address is invalid')
                socket_listen = socket.socket(family, socket.SOCK_DGRAM if self.udp else socket.SOCK_STREAM)
                socket_listen.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                if not self.udp and self.nodelay:
                    socket_listen.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
                if self.reuseport:
                    socket_listen.setsockopt(socket.SOL_SOCKET, 15, 1)                    
                if self.bindaddress is not None:
                    for b in self.bindaddress:
                        if b[0] == family:
                            socket_listen.bind((b[1], self.port))
                            break
                else:
                    if family == socket.AF_INET:
                        socket_listen.bind(('0.0.0.0',self.port))
                    else:
                        socket_listen.bind(('::',self.port))
                self.valid_addresses = None
            if self.udp:
                try:
                    self.socket = socket_listen
                    self.socket.setblocking(False)
                    self.scheduler.registerPolling(self.socket)
                    read_udp = PollEvent.createMatcher(self.socket.fileno(), PollEvent.READ_READY)
                    err_udp = PollEvent.createMatcher(self.socket.fileno(), PollEvent.ERROR)
                    connected = False
                    while not connected:
                        yield (read_udp, err_udp)
                        if self.event.category == PollEvent.READ_READY:
                            while not connected:
                                try:
                                    data, remote_addr = self.socket.recvfrom(65536, socket.MSG_PEEK)
                                    self.logger.debug('Udp socket receive data from %s', remote_addr)
                                except socket.error as exc:
                                    if exc.args[0] == errno.EWOULDBLOCK or exc.args[0] == errno.EAGAIN:
                                        break
                                    else:
                                        raise
                                if self.valid_addresses is not None:
                                    if remote_addr[0] not in self.valid_addresses:
                                        self.socket.recvfrom(1)
                                    else:
                                        connected = True
                                else:
                                    connected = True
                        elif self.event.category == PollEvent.ERROR or self.event.category == PollEvent.HANGUP:
                            raise IOError('Listen socket is closed')
                    try:
                        err = self.socket.connect_ex(remote_addr)
                        if err == errno.EINPROGRESS:
                            connect_match = PollEvent.createMatcher(self.socket.fileno())
                            for m in self.waitWithTimeout(self.connect_timeout, connect_match):
                                yield m
                            if self.timeout:
                                raise IOError('timeout')
                            else:
                                err = self.socket.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
                                if err != 0:
                                    raise IOError('socket error: ' + str(err))
                        elif err != 0:
                            raise IOError('socket error: ' + str(err))
                    except:
                        self.logger.debug('Failed to connect to remote address: ' + repr(addr), exc_info = True)
                        self.scheduler.unregisterPolling(self.socket)
                        self.socket.close()
                        self.socket = None
                        raise
                finally:
                    if self.unix:
                        try:
                            os.remove(self.path)
                        except Exception:
                            pass
            else:
                try:
                    socket_listen.setblocking(False)
                    self.scheduler.registerPolling(socket_listen, POLLING_IN)
                    socket_listen.listen(8)
                    m = PollEvent.createMatcher(socket_listen.fileno())
                    connected = False
                    while not connected:
                        yield (m,)
                        if self.event.category == PollEvent.READ_READY:
                            while not connected:
                                try:
                                    self.socket, remote_addr = socket_listen.accept()
                                except socket.error as exc:
                                    if exc.args[0] == errno.EWOULDBLOCK or exc.args[0] == errno.EAGAIN:
                                        break
                                    else:
                                        raise
                                if self.valid_addresses is not None:
                                    if remote_addr[0] not in self.valid_addresses:
                                        self.socket.close()
                                    else:
                                        connected = True
                                else:
                                    connected = True
                        elif self.event.category == PollEvent.ERROR and self.event.category == PollEvent.HANGUP:
                            raise IOError('Listen socket is closed')
                finally:
                    self.scheduler.unregisterPolling(socket_listen)
                    socket_listen.close()
                    if self.unix:
                        try:
                            os.remove(self.path)
                        except Exception:
                            pass
                self.socket.setblocking(False)
                self.scheduler.registerPolling(self.socket)
        else:
            if self.unix:
                family = socket.AF_UNIX
                self.addresses = (self.path,)
            else:
                if self.bindaddress is not None:
                    bind_protos = set(b[0] for b in self.bindaddress)
                else:
                    bind_protos = set((socket.AF_INET,socket.AF_INET6))
                addrinfo_protos = set(addr[0] for addr in self.addrinfo)
                if socket.AF_INET in bind_protos and socket.AF_INET in addrinfo_protos:
                    family = socket.AF_INET
                elif socket.AF_INET6 in bind_protos and socket.AF_INET6 in addrinfo_protos:
                    family = socket.AF_INET6
                else:
                    raise ValueError('Target address has different address family with local bind address')
                self.addresses = tuple(addr[4] for addr in self.addrinfo if addr[0] == family)
            for addr in self.addresses:
                self.socket = None
                try:
                    self.socket = socket.socket(family, socket.SOCK_DGRAM if self.udp else socket.SOCK_STREAM)
                except:
                    self.logger.debug('Failed to create socket for family: ' + repr(family), exc_info = True)
                    continue
                if not self.unix and not self.udp and self.nodelay:
                    self.socket.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
                try:
                    if self.bindaddress is not None:
                        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                        for b in self.bindaddress:
                            if b[0] == family:
                                self.logger.debug('Bind socket to address: %s', b[1])
                                self.socket.bind(b[1])
                                break
                    elif self.udp:
                        if family == socket.AF_INET:
                            self.socket.bind(('0.0.0.0',0))
                        elif family == socket.AF_INET6:
                            self.socket.bind(('::',0))
                    self.socket.setblocking(False)
                    self.scheduler.registerPolling(self.socket)
                except:
                    self.logger.debug('Failed to bind to local address: ' + repr(self.bindaddress), exc_info = True)
                    self.socket.close()
                    self.socket = None
                    continue
                try:
                    err = self.socket.connect_ex(addr)
                    if err == errno.EINPROGRESS or err == errno.EWOULDBLOCK or err == errno.EAGAIN:
                        connect_match = PollEvent.createMatcher(self.socket.fileno())
                        for m in self.waitWithTimeout(self.connect_timeout, connect_match):
                            yield m
                        if self.timeout:
                            raise Exception('timeout')
                        else:
                            err = self.socket.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
                            if err == 0:
                                break
                            else:
                                raise Exception('socket error: ' + errno.errorcode.get(err, str(err)))
                    elif err != 0:
                        raise Exception('socket error: ' + errno.errorcode.get(err, str(err)))
                    else:
                        break
                except:
                    self.logger.debug('Failed to connect to remote address: ' + repr(addr), exc_info = True)
                    self.scheduler.unregisterPolling(self.socket)
                    self.socket.close()
                    self.socket = None
                    continue
        if self.socket is None:
            raise IOError('Cannot create connection')
        if self.ssl:
            try:
                self.logger.debug('Wrapping socket with SSL Context')
                self.socket = ssl.wrap_socket(self.socket, self.key, self.certificate, self.passive, ssl.CERT_NONE if self.ca_certs is None else ssl.CERT_REQUIRED,
                                              self.sslversion, self.ca_certs, False)
                handshake = False
                read_matcher = PollEvent.createMatcher(self.socket.fileno(), PollEvent.READ_READY)
                write_matcher = PollEvent.createMatcher(self.socket.fileno(), PollEvent.WRITE_READY)
                error_matcher = PollEvent.createMatcher(self.socket.fileno(), PollEvent.ERROR)
                while not handshake:
                    try:
                        self.logger.debug('Doing handshake')
                        self.socket.do_handshake()
                        handshake = True
                    except ssl.SSLError as exc:
                        if exc.args[0] == ssl.SSL_ERROR_WANT_READ:
                            yield (read_matcher, error_matcher)
                            if self.matcher is error_matcher:
                                raise IOError('Socket closed or get error status before SSL handshake complete')
                        elif exc.args[0] == ssl.SSL_ERROR_WANT_WRITE:
                            yield (write_matcher, error_matcher)
                            if self.matcher is error_matcher:
                                raise IOError('Socket closed or get error status before SSL handshake complete')
                        else:
                            raise
            except:
                self.scheduler.unregisterPolling(self.socket)
                self.socket.close()
                self.socket = None
                raise
    def _reconnect_internal(self):
        timeretry = 0
        for timewait in self.reconnect_timeseq():
            for m in self.waitWithTimeout(timewait):
                yield m
            try:
                with closing(self.create_socket()) as cs:
                    for m in cs:
                        yield m
                break
            except IOError as exc:
                timeretry += 1
                if timeretry > 3:
                    self.logger.warning('Reconnect failed after %d times retry, url=%s', timeretry, self.rawurl, exc_info = True)
    def _reconnect(self):
        matcher = ConnectionControlEvent.createMatcher(self, ConnectionControlEvent.SHUTDOWN, _ismatch = lambda x: x.connmark == self.connmark or x.connmark < 0)
        try:
            self.connected = False
            with closing(self.withException(self._reconnect_internal(), matcher)) as g:
                for m in g:
                    yield m
            self.connected = True
        except RoutineException:
            self.need_reconnect = False
            self.connected = False
    def main(self):
        self.connmark = -1
        try:
            matcher = ConnectionControlEvent.createMatcher(self, ConnectionControlEvent.SHUTDOWN, _ismatch = lambda x: x.connmark == self.connmark or x.connmark < 0)
            with closing(self.withException(self.create_socket(), matcher)) as g:
                for m in g:
                    yield m
        except RoutineException:
            return
        except IOError:
            self.logger.warning('Connection failed for url: %s', self.rawurl, exc_info = True)
            if self.need_reconnect:
                with closing(self._reconnect()) as g:
                    for m in g:
                        yield m
                if not self.connected:
                    for m in self.protocol.notconnected(self):
                        yield m
            else:
                for m in self.protocol.notconnected(self):
                    yield m
        if self.socket:
            with closing(Connection.main(self)) as g:
                for m in g:
                    yield m
    def __repr__(self, *args, **kwargs):
        return Connection.__repr__(self, *args, **kwargs) + '(url=' + self.rawurl + ')'

class TcpServer(RoutineContainer):
    '''
    A server receiving multiple connections
    '''
    logger = logging.getLogger(__name__ + '.TcpServer')
    def __init__(self, url, protocol, scheduler = None, key = None, certificate = None, ca_certs = None):
        RoutineContainer.__init__(self, scheduler)
        self.protocol = protocol
        self.rawurl = url
        self.url = urlsplit(url, 'tcp')
        if self.url.scheme == 'unix' or self.url.scheme == 'lunix':
            self.udp = False
            self.passive = True
            self.ssl = False
            self.unix = True
        elif self.url.scheme == 'ssl' or self.url.scheme == 'lssl':
            self.udp = False
            self.passive = True
            self.ssl = True
            self.unix = False
        else:
            self.udp = False
            self.passive = True
            self.ssl = False
            self.unix = False
        if self.unix and not self.url.path:
            raise ValueError('Unix socket path is not specified in url: ' + url)
        if not self.unix:
            self.hostname = self.url.hostname
            if not self.url.port:
                self.port = self.protocol.defaultport
            else:
                self.port = self.url.port
        else:
            self.path = self.url.path
        self.key = key
        self.certificate = certificate
        self.ca_certs = ca_certs
        self.retry_listen = getattr(self.protocol, 'listen_persist', True)
        self.retry_interval = getattr(self.protocol, 'retrylisten_interval', 3)
        if self.ssl:
            self.sslversion = getattr(self.protocol, 'sslversion', ssl.PROTOCOL_SSLv23)
        self.backlogsize = getattr(self.protocol, 'backlogsize', 2048)
        self.reuseport = getattr(self.protocol, 'reuseport', False)
        self.nodelay = getattr(self.protocol, 'nodelay', False)
        self.logger = ContextAdapter(self.logger, {'context':{'server':self, 'url':self.rawurl, 'protocol':protocol}})
        # Counters
        self.totalaccepts = 0
        self.accepts = 0
        self.listening = False
    def _connection(self, newsock, newproto):
        try:
            newsock.setblocking(False)
            if self.nodelay:
                newsock.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
            self.scheduler.registerPolling(newsock, POLLING_IN|POLLING_OUT)
            if self.ssl:
                handshake = False
                read_matcher = PollEvent.createMatcher(newsock.fileno(), PollEvent.READ_READY)
                write_matcher = PollEvent.createMatcher(newsock.fileno(), PollEvent.WRITE_READY)
                error_matcher = PollEvent.createMatcher(newsock.fileno(), PollEvent.ERROR)
                while not handshake:
                    try:
                        self.logger.debug('Doing handshake on accepted socket: %d', newsock.fileno())
                        newsock.do_handshake()
                        handshake = True
                    except ssl.SSLError as exc:
                        if exc.args[0] == ssl.SSL_ERROR_WANT_READ:
                            yield (read_matcher, error_matcher)
                            if self.matcher is error_matcher:
                                raise IOError('Socket closed or get error status before SSL handshake complete')
                        elif exc.args[0] == ssl.SSL_ERROR_WANT_WRITE:
                            yield (write_matcher, error_matcher)
                            if self.matcher is error_matcher:
                                raise IOError('Socket closed or get error status before SSL handshake complete')
                        else:
                            raise
            conn = Connection(newproto, newsock, self.scheduler)
            conn.need_reconnect = False
            conn.start()
            self.accepts += 1
            self.totalaccepts += 1
        except:
            self.scheduler.unregisterPolling(newsock)
            newsock.close()
            raise
    def _server(self):
        if not self.unix:
            request = (None if not self.hostname else self.hostname, self.port, socket.AF_UNSPEC, socket.SOCK_STREAM, socket.IPPROTO_TCP, socket.AI_ADDRCONFIG|socket.AI_PASSIVE)
            # Resolve hostname
            for m in self.waitForSend(ResolveRequestEvent(request)):
                yield m
            for m in self.waitWithTimeout(20, ResolveResponseEvent.createMatcher(request)):
                yield m
            if self.timeout:
                # Resolve is only allowed through asynchronous resolver 
                self.addrinfo = socket.getaddrinfo(self.hostname, self.port, socket.AF_UNSPEC, socket.SOCK_DGRAM if self.udp else socket.SOCK_STREAM, socket.IPPROTO_UDP if self.udp else socket.IPPROTO_TCP, socket.AI_ADDRCONFIG|socket.AI_NUMERICHOST)
            else:
                if hasattr(self.event, 'error'):
                    raise IOError('Cannot resolve hostname: ' + self.hostname)
                self.addrinfo = self.event.response
        if self.unix:
            self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            if self.reuseport:
                self.socket.setsockopt(socket.SOL_SOCKET, 15, 1)
            self.socket.bind(self.path)
            family = socket.AF_UNIX
        else:
            addrinfo_protos = set(addr[0] for addr in self.addrinfo)
            if socket.AF_INET in addrinfo_protos:
                family = socket.AF_INET
            elif socket.AF_INET6 in addrinfo_protos:
                family = socket.AF_INET6
            else:
                raise ValueError('Local bind address is invalid')
            self.socket = socket.socket(family, socket.SOCK_STREAM)
            if self.nodelay:
                self.socket.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            if self.reuseport:
                self.socket.setsockopt(socket.SOL_SOCKET, 15, 1)
            for b in self.addrinfo:
                if b[0] == family:
                    self.socket.bind(b[4])
                    break
        try:
            self.socket.setblocking(False)
            if self.ssl:
                self.socket = ssl.wrap_socket(self.socket, self.key, self.certificate, True, ssl.CERT_NONE if self.ca_certs is None else ssl.CERT_REQUIRED,
                                              self.sslversion, self.ca_certs, False)                
            for m in self.protocol.beforelisten(self, self.socket):
                yield m
            self.scheduler.registerPolling(self.socket, POLLING_IN)
            self.socket.listen(self.backlogsize)
            self.listening = True
            try:
                self.localaddr = self.socket.getsockname()
            except:
                pass
            m = PollEvent.createMatcher(self.socket.fileno())
            while True:
                yield (m,)
                if self.event.category == PollEvent.READ_READY:
                    while True:
                        try:
                            new_socket, remote_addr = self.socket.accept()
                        except socket.error as exc:
                            if exc.errno == errno.EWOULDBLOCK or exc.errno == errno.EAGAIN:
                                break
                            else:
                                self.logger.warning('Unexpected exception on accepting', exc_info = True)
                                break
                        try:
                            if hasattr(self.protocol, 'accept'):
                                new_proto = self.protocol.accept(self, remote_addr, new_socket)
                                if new_proto is None:
                                    new_socket.close()
                                    continue
                            else:
                                new_proto = self.protocol
                            self.subroutine(self._connection(new_socket, new_proto)) 
                        except:
                            new_socket.close() 
                else:
                    self.logger.warning('Error polling status received: ' + repr(self.event))
                    break
        finally:
            self.scheduler.unregisterPolling(self.socket)
            self.socket.close()
            if self.unix:
                try:
                    os.remove(self.path)
                except Exception:
                    pass
            self.listening = False
    def main(self):
        self.connmark = 0
        matcher = ConnectionControlEvent.createMatcher(self, ConnectionControlEvent.SHUTDOWN, _ismatch = lambda x: x.connmark == self.connmark or x.connmark < 0)
        matcher2 = ConnectionControlEvent.createMatcher(self, ConnectionControlEvent.STOPLISTEN, _ismatch = lambda x: x.connmark == self.connmark or x.connmark < 0)
        matcher3 = ConnectionControlEvent.createMatcher(self, ConnectionControlEvent.STARTLISTEN, _ismatch = lambda x: x.connmark == self.connmark or x.connmark < 0)
        retry = True
        self.retry_interval = None
        try:
            while retry:
                self.connmark += 1
                self.accepts = 0
                try:
                    with closing(self.withException(self._server(), matcher, matcher2)) as g:
                        for m in g:
                            yield m
                except RoutineException:
                    if self.matcher is matcher:
                        retry = False
                    else:
                        yield (matcher, matcher3)
                        if self.matcher is matcher:
                            retry = False
                        else:
                            retry = True
                except IOError:
                    retry = self.retry_listen
                    if retry:
                        self.logger.warning('Begin listen failed on URL: %s', self.rawurl, exc_info = True)
                        for m in self.waitWithTimeout(self.retry_interval):
                            yield m
                    else:
                        raise
        finally:
            self.subroutine(self._final())
    def shutdown(self, connmark = -1):
        '''
        Can call without delegate
        '''
        if connmark is None:
            connmark = self.connmark
        for m in self.waitForSend(ConnectionControlEvent(self, ConnectionControlEvent.SHUTDOWN, True, connmark)):
            yield m
    def stoplisten(self, connmark = -1):
        '''
        Can call without delegate
        '''
        if connmark is None:
            connmark = self.connmark
        for m in self.waitForSend(ConnectionControlEvent(self, ConnectionControlEvent.STOPLISTEN, True, connmark)):
            yield m    
    def startlisten(self, connmark = -1):
        '''
        Can call without delegate
        '''
        if connmark is None:
            connmark = self.connmark
        for m in self.waitForSend(ConnectionControlEvent(self, ConnectionControlEvent.STARTLISTEN, True, connmark)):
            yield m    
    def _final(self):
        for m in self.protocol.serverfinal(self):
            yield m
    def __repr__(self, *args, **kwargs):
        baserepr = RoutineContainer.__repr__(self, *args, **kwargs)
        return baserepr + '(Listen on %r)' % (getattr(self, 'localaddr', None),)



#! /usr/bin/python
#! --*-- utf-8 --*--
import logging
from uuid import uuid1

from vlcp.server.module import Module,api,publicapi
from vlcp.event.runnable import RoutineContainer
from vlcp.utils.dataobject import updater,set_new,ReferenceObject,dump
from vlcp.utils.networkmodel import *
from vlcp.utils.ethernet import ETHERTYPE_8021Q 
from vlcp.config.config import defaultconfig

logger = logging.getLogger('NetworkVlanDriver')

@defaultconfig
class NetworkVlanDriver(Module):
    """
    Network driver for VXLAN networks. When creating a VXLAN type physical network,
    you must specify an extra option ``vlanrange``.
    """
    def __init__(self,server):
        super(NetworkVlanDriver,self).__init__(server)
        self.app_routine = RoutineContainer(self.scheduler)
        self.app_routine.main = self._main
        self.routines.append(self.app_routine)
        self.createAPI(
                       publicapi(self.createphysicalnetworks,
                                    criteria=lambda networks,type:type == 'vlan'),
                       publicapi(self.updatephysicalnetworks,
                                    criteria=lambda type,networks:type == 'vlan'),
                       publicapi(self.deletephysicalnetworks,
                                    criteria=lambda type,networks:type == 'vlan'),
                       publicapi(self.createphysicalports,
                                    criteria=lambda type,ports:type == 'vlan'),
                       publicapi(self.updatephysicalports,
                                    criteria=lambda phynettype,
                                    ports:phynettype == 'vlan'),
                       publicapi(self.deletephysicalports,
                                    criteria=lambda phynettype,
                                    ports:phynettype == 'vlan'),
                       publicapi(self.createlogicalnetworks,
                                    criteria=lambda phynettype,
                                    networks:phynettype == 'vlan'),
                       publicapi(self.updatelogicalnetworks,
                                    criteria=lambda phynettype,
                                    networks:phynettype == 'vlan'),
                       publicapi(self.deletelogicalnetworks,
                                    criteria=lambda phynettype,networks:phynettype == "vlan"),
                       #used in IOprocessing module
                       publicapi(self.createioflowparts,
                                    criteria=lambda connection,logicalnetwork,
                                    physicalport,logicalnetworkid,physicalportid:
                                    logicalnetwork.physicalnetwork.type == "vlan")

                       )

    def _main(self):

        logger.info("network_vlan_driver running ---")
        if None:
            yield

    def createphysicalnetworks(self,networks,type):
        
        new_networks = [ self._createphysicalnetwork(**n) for n in networks]
        
        def createphynetworks(keys,values):
            for i in range(0,len(new_networks)): 
                values[i + 1] = set_new(values[i + 1],new_networks[i][0])
                values[i + 1 + len(new_networks)] = set_new(values[i + 1 + len(new_networks)],
                        new_networks[i][1])
                values[0].set.dataset().add(new_networks[i][0].create_weakreference())

            return keys,values

        return createphynetworks
    
    def _createphysicalnetwork(self,type,id,**args):

        if 'vlanrange' not in args:
            raise ValueError('must specify vlanrange with network type vlan')

        
        #
        # vlanrange [(1,100),(200,500)]
        #
        try:
            lastend = 0
            for start,end in args.get('vlanrange'):
                if start > end or start <= lastend:
                    raise ValueError('vlan sequences overlapped or disorder')
                if end > 4095:
                    raise ValueError('vlan out of range (0 -- 4095)')  
                lastend = end
        except:
            raise ValueError('vlanrange format error,[(1,100),(200,500)]')


        # create an new physical network
        new_network = PhysicalNetwork.create_instance(id)
        new_network.type = type

        for k,v in args.items():
            setattr(new_network,k,v)
        
        # create 1 : 1 physical network map
        new_networkmap = PhysicalNetworkMap.create_instance(id)
        new_networkmap.network = new_network.create_weakreference()
        
        return new_network,new_networkmap

    
    def updatephysicalnetworks(self,type,networks):
        def updatephynetworks(keys,values):
            phykeys = keys[0:len(keys)//2]
            phyvalues = values[0:len(values)//2]
            
            phymapkeys = keys[len(keys)//2:]
            phymapvalues = values[len(keys)//2:]

            phynetdict = dict(zip(phykeys,zip(phyvalues,phymapvalues)))

            for network in networks:
                phynet,phymap = phynetdict.get(PhysicalNetwork.default_key(network.get('id')))
                if not phynet or not phymap:
                    raise ValueError("key object not existed "+\
                            PhysicalNetwork.default_key(network['id']))
                
                if 'vlanrange' in network:
                    findflag = False
                    for k,_ in phymap.network_allocation.items():
                        find = False
                        for start,end in network.get('vlanrange'):
                            if int(k) >= start and int(k) <= end:
                                find = True
                                break
                        if find == False:
                            findflag = False
                            break
                        else:
                            findflag = True
                    
                    if len(phymap.network_allocation) == 0:
                        findflag = True
                    
                    # new vlanrange do not 
                    if findflag == False:
                        raise ValueError('new vlan range do not match with allocation')
                for k,v in network.items():   
                    setattr(phynet,k,v)
            return phykeys,phyvalues

        return updatephynetworks
    
    def deletephysicalnetworks(self,type,networks):

        def deletephynetwork(keys,values):
            
            phynetlen = (len(keys) - 1)//2
            phynetkeys = keys[1:1+phynetlen]
            phynetvalues = values[1:1+phynetlen]

            phynetmapkeys = keys[1+phynetlen:]
            phynetmapvalues = values[1+phynetlen:]

            phynetdict = dict(zip(phynetkeys,zip(phynetvalues,phynetmapvalues)))
            phynetset = values[0].set.dataset()
            for network in networks:
                phynet,phynetmap = phynetdict.get(PhysicalNetwork.default_key(network.get('id')))
                # if there is phynetworkport on the phynet 
                # delete will fail
                if phynetmap and phynetmap.ports.dataset():
                    raise ValueError("delete all phynetworkport on this phynet before delete")
                # if there is logicnetwork on the phynet
                # delete will fail
                if phynetmap and phynetmap.network_allocation:
                    raise ValueError('delete all logicnetwork on this phynet before delete')
                
                phynetset.discard(phynet.create_weakreference())
           
            return keys,[values[0]]+[None]*(len(keys)-1)
        
        return deletephynetwork
    def createphysicalports(self,type,ports):
        portobjs = [self._createphysicalport(**n) for n in ports]
        
        def createpyports(keys,values):
            phynetlen = (len(keys) - 1 - len(portobjs))//2
            
            phynetkeys = keys[1+len(portobjs):1+len(portobjs)+phynetlen]
            phynetvalues = values[1+len(portobjs):1+len(portobjs)+phynetlen]
            
            phynetmapkeys = keys[1+len(portobjs)+phynetlen:]
            phynetmapvalues = values[1+len(portobjs)+phynetlen:]
            
            phynetdict = dict(zip(phynetkeys,zip(phynetvalues,phynetmapvalues)))

            for i in range(0,len(portobjs)):
                values[i + 1] = set_new(values[i + 1],portobjs[i])
                
                key = portobjs[i].physicalnetwork.getkey()
                phynet,phymap = phynetdict.get(key)
                if not phynet or not phymap:
                    raise ValueError("key object not existed "+ key) 

                phymap.ports.dataset().add(portobjs[i].create_weakreference())
                values[0].set.dataset().add(portobjs[i].create_weakreference())

            return keys[0:1+len(ports)] + phynetmapkeys,values[0:1+len(ports)] + phynetmapvalues
        return createpyports
    def _createphysicalport(self,physicalnetwork,name,vhost,systemid,bridge,**args):

        p = PhysicalPort.create_instance(vhost,systemid,bridge,name)
        p.physicalnetwork = ReferenceObject(PhysicalNetwork.default_key(physicalnetwork))
        
        for k,v in args.items():
            setattr(p,k,v)

        return p
    
    def updatephysicalports(self,phynettype,ports):

        def updatephyports(keys,values):
            
            portdict = dict(zip(keys,values))
            for port in ports:
                
                key = PhysicalPort.default_key(port['vhost'],port['systemid'],
                            port['bridge'],port['name'])
                phyport = portdict[key]
                if phyport:
                    for k,v in port.items():
                        setattr(phyport,k,v)
                else:
                    raise ValueError("key object not existed "+ key) 
            
            return keys,values

        return updatephyports
   
    def deletephysicalports(self,phynettype,ports):

        def deletephyports(keys,values):
            portkeys = keys[1:1+len(ports)]
            portvalues = values[1:1+len(ports)]
            
            phynetmapkeys = keys[1+len(ports):]
            phynetmapvalues = values[1+len(ports):]
            
            phynetmapdict = dict(zip(phynetmapkeys,phynetmapvalues))
            portdict = dict(zip(portkeys,portvalues))

            for port in ports:
                phymap = phynetmapdict.get(PhysicalNetworkMap.default_key(port.get("phynetid")))
                
                key = PhysicalPort.default_key(port['vhost'],port['systemid'],port['bridge'],port['name'])
                portobj = portdict.get(key)
                
                if not portobj or not phymap:
                    raise ValueError("key object not existed "+ key) 

                phymap.ports.dataset().discard(portobj.create_weakreference())
               
                values[0].set.dataset().discard(portobj.create_weakreference())
           
            return keys,[values[0]]+[None]*len(ports) + phynetmapvalues
        return deletephyports
    def createlogicalnetworks(self,phynettype,networks):
        
        networkmap = [self._createlogicalnetwork(**n) for n in networks]

        def createlgnetworks(keys,values):
            
            phynetlen = (len(keys) - len(networkmap)*2 - 1)//2
            phynetkeys = keys[1 + len(networkmap)*2 : 1 + len(networkmap)*2 + phynetlen]
            phynetvalues = values[1 + len(networkmap)*2 : 1 + len(networkmap)*2 + phynetlen]
            
            phynetmapkeys = keys[1+len(networkmap)*2 + phynetlen:]
            phynetmapvalues = values[1+len(networkmap)*2 + phynetlen:]
            
            phynetmapdict = dict(zip(phynetkeys,zip(phynetvalues,phynetmapvalues)))
            
            #
            # have an problem ,  when [{...},{... vlanid is least or last+1}]
            # first one will allocate least one vlanid,second will conflict  
            #
            # so set new lgnet that have user defind 'vlanid' first
            for i in range(0,len(networks)):
                if hasattr(networkmap[i][0],'vlanid'):

                    vlanid = int(networkmap[i][0].vlanid)
                    networkmap[i][0].vlanid = vlanid
                    phynet,phymap = phynetmapdict.get(networkmap[i][0].physicalnetwork.getkey())

                    if not phynet or not phymap:
                        raise ValueError("physicalnetwork key object not existed "+\
                                networkmap[i][0].physicalnetwork.getkey()) 

                    if _isavaliablevlanid(phynet.vlanrange,phymap.network_allocation.keys(),vlanid):
                        phymap.network_allocation[str(vlanid)] = networkmap[i][0].create_weakreference()
                    else:
                        raise ValueError("user defind vlan id has been used or out of range!")
                    
                    # set lgnetwork
                    values[1+i] = set_new(values[i + 1],networkmap[i][0])
                    # set lgnetworkmap
                    values[1+i+len(networks)] = set_new(values[i + 1 + len(networks)],networkmap[i][1])
                    # set phynetmap
                  
                    #_,phymap = phynetmapdict.get(networkmap[i][0].physicalnetwork.getkey())
                    phymap.logicnetworks.dataset().add(networkmap[i][0].create_weakreference())

                    values[0].set.dataset().add(networkmap[i][0].create_weakreference())


            for i in range(0,len(networks)):
                if not getattr(networkmap[i][0],'vlanid',None):
                    # there is no 'vlanid' in lgnetwork
                    # allocated one from vlanrange

                    phynet,phymap = phynetmapdict.get(networkmap[i][0].physicalnetwork.getkey())
                    
                    if not phynet or not phymap:
                        raise ValueError("physicalnetwork key object not existed "+\
                                networkmap[i][0].physicalnetwork.getkey()) 

                    vlanid = _findavaliablevlanid(phynet.vlanrange,phymap.network_allocation.keys())
                    if not vlanid:
                        raise ValueError("there is no avaliable vlan id")
                    setattr(networkmap[i][0],'vlanid',vlanid)
                    phymap.network_allocation[str(vlanid)] = networkmap[i][0].create_weakreference()
                    
                    # set lgnetwork
                    values[1+i] = set_new(values[i + 1],networkmap[i][0])
                    # set lgnetworkmap
                    values[1+i+len(networks)] = set_new(values[i + 1 + len(networks)],networkmap[i][1])
                    # set phynetmap
                  
                    #_,phymap = phynetmapdict.get(networkmap[i][0].physicalnetwork.getkey())
                    phymap.logicnetworks.dataset().add(networkmap[i][0].create_weakreference())

                    values[0].set.dataset().add(networkmap[i][0].create_weakreference())
            return keys[0:1] + keys[1:1+len(networks)+len(networks)]+phynetmapkeys,\
                    values[0:1]+values[1:1+len(networks)+len(networks)] + phynetmapvalues
        return createlgnetworks
        
    def _createlogicalnetwork(self,physicalnetwork,id,**args):

        logicalnetwork = LogicalNetwork.create_instance(id)
        logicalnetworkmap = LogicalNetworkMap.create_instance(id)
        
        for k,v in args.items():
            setattr(logicalnetwork,k,v)

        logicalnetworkmap.network = logicalnetwork.create_reference()
        logicalnetwork.physicalnetwork = ReferenceObject(PhysicalNetwork.default_key(physicalnetwork))

        return (logicalnetwork,logicalnetworkmap)
    
   
    def updatelogicalnetworks(self,phynettype,networks):

        def updatelgnetworks(keys,values):
            
            phynetlen = (len(keys) - len(networks))//2
            
            lgnetkeys = keys[0:len(networks)]
            lgnetvalues = values[0:len(networks)]
            phynetkeys = keys[len(networks):len(keys)+phynetlen]
            phynetvalues = values[len(networks):len(keys)+phynetlen]
            
            phynetmapkeys = keys[len(networks)+phynetlen:]
            phynetmapvalues = values[len(networks)+phynetlen:]

            phynetdict = dict(zip(phynetkeys,zip(phynetvalues,phynetmapvalues)))
            lgnetdict = dict(zip(lgnetkeys,lgnetvalues))
            
            for network in networks:
                # 
                # to update vlanid, we should delete it first ,
                #
                if "vlanid" in network:
                    phynet,phynetmap = phynetdict.get(PhysicalNetwork.default_key(network.get("phynetid")))
                    
                    if not phynet or not phynetmap:
                        raise ValueError("physicalnetwork key object not existed "+\
                            PhysicalNetwork.default_key(network["phynetid"])) 
                    
                    lgnet = lgnetdict.get(LogicalNetwork.default_key(network["id"]))
                    
                    del phynetmap.network_allocation[str(lgnet.vlanid)]

            for network in networks:
                phynet,phynetmap = phynetdict.get(PhysicalNetwork.default_key(network.get("phynetid")))
                if not phynet or not phynetmap:
                    raise ValueError("physicalnetwork key object not existed "+\
                        PhysicalNetwork.default_key(network["phynetid"])) 
                
                lgnet = lgnetdict.get(LogicalNetwork.default_key(network["id"]))
               
                if "vlanid" in network:
                    vlanid = int(network["vlanid"])
                    
                    if vlanid == lgnet.vlanid:
                        continue

                    if _isavaliablevlanid(phynet.vlanrange,phynetmap.network_allocation.keys(),vlanid):
                        phynetmap.network_allocation[str(vlanid)] = lgnet.create_weakreference()
                    else:
                        raise ValueError("new vlanid is not avaliable")
                    
                    setattr(lgnet,'vlanid',vlanid)
                
                for k,v in network.items():
                    # this phynetid is fack attr for find phynet phymap
                    if k != 'phynetid' and k != 'vlanid':
                        setattr(lgnet,k,str(v))
            return keys[0:len(networks)] + phynetmapkeys ,values[0:len(networks)] + phynetmapvalues

        return updatelgnetworks

    def deletelogicalnetworks(self,phynettype,networks):

        def deletelgnetworks(keys,values):
            phynetlen = (len(keys) - 1 - len(networks))//2

            lgnetkeys = keys[1:1+len(networks)]
            lgnetvalues = values[1:1+len(networks)]
            
            lgnetmapkeys = keys[1+len(networks):1+len(networks)+len(networks)]
            lgnetmapvalues = values[1+len(networks):1+len(networks)+len(networks)]
            
            phynetmapkeys = keys[1+len(networks)+len(networks):]
            phynetmapvalues = values[1+len(networks)+len(networks):]

            lgnetdict = dict(zip(lgnetkeys,zip(lgnetvalues,lgnetmapvalues)))
            phynetmapdict = dict(zip(phynetmapkeys,phynetmapvalues))
            
            for network in networks:
                lgnet,lgnetmap = lgnetdict.get(LogicalNetwork.default_key(network.get("id")))

                phymap = phynetmapdict.get(PhysicalNetworkMap.default_key(network.get("phynetid")))
                
                if not phymap:
                    raise ValueError("physicalnetwork map key object not existed " + network["phynetid"]) 

                values[0].set.dataset().discard(lgnet.create_weakreference())
                phymap.logicnetworks.dataset().discard(lgnet.create_weakreference())
                del phymap.network_allocation[str(lgnet.vlanid)]
            
            return keys,[values[0]]+[None]*len(networks)*2+phynetmapvalues
        return deletelgnetworks

    def createioflowparts(self,connection,logicalnetwork,physicalport,logicalnetworkid,physicalportid):

        #
        #  1. used in IOProcessing , when physicalport add to logicalnetwork 
        #     return : input flow match vlan oxm, input flow vlan parts actions
        #              output flow vlan parts actions, output group bucket
        #
        
        input_match_oxm = [
                    connection.openflowdef.create_oxm(
                        connection.openflowdef.OXM_OF_VLAN_VID,
                        logicalnetwork.vlanid|connection.openflowdef.OFPVID_PRESENT)
                ]

        input_action = [
                   connection.openflowdef.ofp_action(type = 
                        connection.openflowdef.OFPAT_POP_VLAN)    
              ]

        output_action = [
                    connection.openflowdef.ofp_action_push(ethertype=ETHERTYPE_8021Q),
                    connection.openflowdef.ofp_action_set_field(
                            field = connection.openflowdef.create_oxm(
                                    connection.openflowdef.OXM_OF_VLAN_VID,
                                    logicalnetwork.vlanid |
                                    connection.openflowdef.OFPVID_PRESENT
                                )
                        ),
                    connection.openflowdef.ofp_action_output(
                            port = physicalportid 
                        )
                ]
        output_action2 = [
                    connection.openflowdef.ofp_action_push(ethertype=ETHERTYPE_8021Q),
                    connection.openflowdef.ofp_action_set_field(
                            field = connection.openflowdef.create_oxm(
                                    connection.openflowdef.OXM_OF_VLAN_VID,
                                    logicalnetwork.vlanid |
                                    connection.openflowdef.OFPVID_PRESENT
                                )
                        ),
                    connection.openflowdef.ofp_action_output(
                            port = connection.openflowdef.OFPP_IN_PORT
                        )
                ]
        
        # this action is same as ouput_action  on type vlan
        output_group_bucket_action = [
                    connection.openflowdef.ofp_action_push(ethertype=ETHERTYPE_8021Q),
                    connection.openflowdef.ofp_action_set_field(
                            field = connection.openflowdef.create_oxm(
                                    connection.openflowdef.OXM_OF_VLAN_VID,
                                    logicalnetwork.vlanid |
                                    connection.openflowdef.OFPVID_PRESENT
                                )
                        ),
                    connection.openflowdef.ofp_action_output(
                            port = physicalportid
                        )
                ]

        return input_match_oxm,input_action,output_action,output_group_bucket_action,output_action2
#
# utils function
#

def _findavaliablevlanid(vlanrange,allocated):
    
    vlanid = None
    for vr in vlanrange:
        find = False
        for v in range(vr[0],vr[1]):
            if str(v) not in allocated:
                vlanid = v
                find = True
                break

        if find:
            break
    return vlanid

def _isavaliablevlanid(vlanrange,allocated,vlanid):
    
    find = False
    for start,end in vlanrange:
        if start <= int(vlanid) <= end:
            find = True
            break

    if find:
        if str(vlanid) not in allocated:
            find = True
        else:
            find = False
    else:
        find = False

    return find
